#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Client tasks (delayed or scheduled)"""

import datetime
from enum import Enum
from typing import Dict, List, Optional, Text, Union

import isodate
from skill_sdk.utils.util import CamelModel


class ReferenceType(Text, Enum):
    """
    Reference event for the execution

        SPEECH_END - end of vocalization of the current response.
        THIS_RESPONSE - immediately after reception of the current response by the touchpoint
        MEDIA_CONTENT_END - end of output of a media content attached to the current response, e.g. an audio stream

    """

    SPEECH_END = "SPEECH_END"
    THIS_RESPONSE = "THIS_RESPONSE"
    MEDIA_CONTENT_END = "MEDIA_CONTENT_END"


class ExecuteAfter(CamelModel):
    """Command is executed after speech end with optional offset"""

    reference: ReferenceType = ReferenceType.SPEECH_END

    # Positive offset relative to reference given as duration
    offset: Optional[Text] = None


class ExecutionTime(CamelModel):
    """
    Exported as timestamp in ISO-8601 format, e.g. 2020-11-25T12:00:00Z
    """

    # Relative execution time
    execute_after: Optional[ExecuteAfter] = None

    # Absolute execution time
    execute_at: Optional[Text] = None

    @staticmethod
    def at(time: datetime.datetime) -> "ExecutionTime":
        return ExecutionTime(execute_at=time.isoformat())

    @staticmethod
    def after(
        event: ReferenceType = ReferenceType.SPEECH_END,
        offset: datetime.timedelta = datetime.timedelta(0),
    ) -> "ExecutionTime":
        return ExecutionTime(
            execute_after=ExecuteAfter(
                reference=event, offset=isodate.duration_isoformat(offset)
            )
        )


class InvokeData(CamelModel):
    """Intent invoke data: name, skill and parameters"""

    # Intent name
    intent: Text

    # Skill Id
    skill_id: Optional[Text] = None

    # Parameters (will be converted to intent invoke attributes)
    parameters: Dict[Text, List[Text]] = {}


class DelayedClientTask(CamelModel):
    """
    Delayed (postponed or scheduled) task, that client executes upon receiving this response
    Standard use case is to invoke an intent after speech end

    """

    # Invoke data
    invoke_data: InvokeData

    # Invoke execution time (default - right after speech end)
    execution_time: ExecutionTime

    @staticmethod
    def invoke(
        intent: Text, skill_id: Text = None, **kwargs: Union[Text, List[Text]]
    ) -> "DelayedClientTask":
        """
        Create a task to invoke intent

            Execute "WEATHER__INTENT" in 10 seconds after speech end:
            >>>         response = Response("Weather forecast for Bonn in 10 seconds.").with_task(
            >>>             ClientTask.invoke("WEATHER__INTENT", location=["Bonn"])
            >>>                 .after(offset=datetime.timedelta(seconds=10))
            >>>         )


        :param intent:      Intent name to invoke
        :param skill_id:    Optional skill Id
        :param kwargs:      Parameters: wey/values map converted into attributes for skill invocation
        :return:
        """
        parameters: Dict[Text, List[Text]] = {
            k: [v] if isinstance(v, Text) else v for k, v in kwargs.items()
        }
        invoke_data = InvokeData(
            intent=intent, skill_id=skill_id, parameters=parameters
        )
        execution_time = ExecutionTime.after(ReferenceType.SPEECH_END)

        return DelayedClientTask(invoke_data=invoke_data, execution_time=execution_time)

    def at(self, time: datetime.datetime) -> "DelayedClientTask":
        """
        Schedule the task execution to particular time point

            Excetute the task in 10 seconds:
            >>> task.at(datetime.datetime.now() + datetime.timedelta(seconds=10))

        :param time:    Time point to execute the task
        :return:
        """
        return self.copy(update=dict(execution_time=ExecutionTime.at(time)))

    def after(
        self,
        event: ReferenceType = ReferenceType.SPEECH_END,
        offset: datetime.timedelta = datetime.timedelta(0),
    ) -> "DelayedClientTask":
        """
        Delay the task execution AND/OR change the reference point type

            Schedule the task to execute BEFORE speech starts:
            >>> task.after(ReferenceType.THIS_RESPONSE)

            To delay task execution by 10 seconds after speech ends:
            >>> task.after(ReferenceType.SPEECH_END, datetime.timedelta(seconds=10))

        :param event:   even reference type (SPEECH_END - after speech ends, THIS_RESPONSE - before speech starts)
        :param offset:  offset timedelta
        :return:
        """
        return self.copy(
            update=dict(execution_time=ExecutionTime.after(event=event, offset=offset))
        )


ClientTask = DelayedClientTask



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill runner"""

import inspect
import logging
from functools import partial
from pathlib import Path
from types import MappingProxyType, ModuleType
from typing import Any, Callable, Dict, Mapping, Text, Union
from fastapi import FastAPI

from skill_sdk import i18n
from skill_sdk.utils import util
from skill_sdk.intents import handlers, invoke
from skill_sdk.responses import Response, ErrorResponse

logger = logging.getLogger(__name__)

#
# Fallback intent: called when no implementation found
#
#   Eq.: a skill handling 'Intent__One' and 'Intent__Two' receives 'Intent__Three' intent.
#
#       Normally this would trigger a "404: intent not found" response.
#
#       If skill implements a handler for 'FALLBACK_INTENT' intent,
#       this function is called instead.
#
FALLBACK_INTENT = "FALLBACK_INTENT"


class Skill(FastAPI):
    """Overloads FastAPI with intent handlers and translations"""

    locales: Mapping[Text, i18n.Translations]

    intents: Mapping[Text, Callable]

    #
    # Temporary dictionary with intent implementations
    #
    __intents: Dict[Text, Callable] = {}

    def __init__(
        self,
        *,
        translations: Dict[Text, i18n.Translations] = None,
        **kwargs,
    ) -> None:

        self.translations = translations or i18n.load_translations()
        self.intents = MappingProxyType(self.__intents)

        super().__init__(**kwargs)

        util.populate_intent_examples(self.intents)

    def get_handler(self, name: Text):
        """
        Return intent handler by intent name
            (or 'FALLBACK_INTENT' handler if not found)

        :param name:    intent name
        :return:
        """
        handler = self.intents.get(name)

        if handler is None:
            logger.debug("Intent %s handler not found.", name)
            handler = self.intents.get(FALLBACK_INTENT)

            if handler is None:
                raise KeyError(
                    f"Intent {name} handler not found and no fallback specified."
                )
            else:
                logger.debug("Using fallback intent handler %s", repr(handler))

        return handler

    def include(
        self,
        intent: Text = None,
        *,
        handler: Callable,
        error_handler: handlers.ErrorHandlerType = None,
    ) -> "Skill":
        """
        Add an intent handler to the skill

        :param intent:          intent name
        :param handler:         intent handler
        :param error_handler:   conversion error handler
        :return:
        """

        if (
            intent is not None
            and intent in self.intents
            and handler is not self.intents[intent]
        ):
            raise ValueError(
                f"Cannot redefine existing intent {repr(intent)} handler: {repr(self.intents[intent])}"
            )

        registered = ""
        try:
            registered = [
                intent
                for intent, implementation in self.intents.items()
                if handler is implementation
            ][0]
            logger.debug(
                "Intent %s handler %s already registered",
                repr(registered),
                repr(handler),
            )
        except IndexError:
            pass

        if intent and not registered:
            self.__intents[intent] = self.__register(intent, handler, error_handler)
            logger.debug("Intent %s handler: %s", repr(intent), repr(handler))

        return self

    def develop(self):
        """Init development mode: load Designer UI"""
        from skill_sdk import ui

        ui.setup(self)
        return self

    @staticmethod
    def __register(
        intent: Text,
        handler: Callable[..., Any],
        error_handler: handlers.ErrorHandlerType = None,
    ):

        if not intent:
            raise ValueError("Intent name is required when adding intent handlers")

        if intent in Skill.__intents:
            raise ValueError(
                f"Duplicate intent {repr(intent)} with handler {repr(handler)}"
            )

        if not inspect.isfunction(handler):
            raise ValueError(
                f"Wrong handler type: {type(handler)}. Expecting coroutine or function."
            )

        decorated = handlers.intent_handler(handler, error_handler=error_handler)
        Skill.__intents[intent] = decorated
        logger.debug("Intent %s static handler: %s", repr(intent), repr(decorated))
        return decorated

    @staticmethod
    def intent_handler(
        intent: Text,
        handler: Callable = None,
        error_handler: handlers.ErrorHandlerType = None,
    ) -> Callable:
        """
        Decorator to wrap an intent implementation

        :param intent:          Intent name
        :param handler:         Handler function
        :param error_handler:   Optional handler to call if conversion error occurs
        :return:
        """

        if callable(intent) and handler is None:
            # If used as simple decorator, i.e.:
            #
            #   @intent_handler
            #   async def handler():
            #
            # In this case we simply return decorated function
            return handlers.intent_handler(intent, error_handler=error_handler)

        return partial(Skill.__register, intent, error_handler=error_handler)

    async def test_intent(
        self,
        intent: Text,
        translation: i18n.Translations = None,
        session: Union[util.CamelModel, Dict[Text, Text]] = None,
        **kwargs,
    ) -> Union[Response, ErrorResponse]:
        """
        Test an intent implementation

        :param intent:      Intent name
        :param translation: Translations to use (NullTranslations if not set)
        :param session:     Mocked session (can be a session result of previous test invoke)
        :param kwargs:      Intent's attributes
        :return:
        """
        handler = self.get_handler(intent)

        r = util.create_request(intent, session=session, **kwargs).with_translation(
            translation if translation else i18n.Translations()
        )

        return await invoke(handler, r)

    @property
    def module(self) -> ModuleType:
        try:
            return getattr(self, "_module")
        except AttributeError:
            raise RuntimeError(
                '"module" property is only available in DEVELOPMENT mode.'
            ) from None

    def reload(self, app_str: Text = "") -> "Skill":
        self.close()

        logger.info(
            "Reloading %s from %s",
            repr(app_str) if app_str else "default app",
            repr(self.module),
        )
        util.reload_recursive(self.module)

        # Get intents and handlers from the application object,
        # if not supplied, the default app is self anyway
        app = getattr(self.module, app_str, None)
        if app is not None:
            self.intents = app.intents

        logger.info("Loaded handlers: %s", list(self.intents))
        return self

    @staticmethod
    def close():
        """
        Cleanup: Skill.__intents is static to enable backward-compatible "@intent_handler" decorators,
        that are not bound to a skill instance (yet).

        While it's not a big issue for a single-purpose skill, it is for SDK unit testing,
        when skill instances dynamically created and destroyed.

        To force cleanup:

            >>> from contextlib import closing
            >>> with closing(init_app()):
            >>>     ... # do your things

        """

        Skill.__intents.clear()


def init_app(
    config_path: Union[Dict, Path, Text] = None,
    develop: bool = None,
    configure_logging: bool = None,
) -> Skill:
    """
    Create FastAPI application from configuration file

    `skill` section of the config file is used to initialize FastAPI app,
    so any FastAPI initialization parameters, like `debug`, `title`, `description`, `version` can be used

    :param config_path:
    :param develop:             Flag to init an app in "development" mode:
                                overrides debug flag from configuration,
                                initializes Designer UI
    :param configure_logging:   If logging settings should be re-initialized:
                                logging is configured within CLI,
                                but sometimes you'd want to re-initialize it explicitly,
                                for example, when using UvicornWorker with Gunicorn

    :return:
    """
    from skill_sdk import config, middleware, log, routes

    # Reload setting from a config file
    config.settings.reload(conf_file=config_path)

    if develop is None:
        develop = config.settings.debug()

    if configure_logging:
        log.setup_logging()

    app_config = {**config.settings.app_config(), **dict(debug=develop)}
    logger.debug("App config: %s", app_config)

    app = Skill(**app_config)

    middleware.setup_middleware(app)
    routes.setup_routes(app)

    return app.develop() if develop else app


intent_handler = Skill.intent_handler


def test_intent(
    intent: Text,
    translation: i18n.Translations = None,
    session: Union[util.CamelModel, Dict[Text, Text]] = None,
    **kwargs,
) -> Union[Response, ErrorResponse]:
    """
    Backward compatible test helper

    :param intent:      Intent name
    :param translation: Translations to use (NullTranslations if not set)
    :param session:     Mocked session (can be a session result of previous test invoke)
    :param kwargs:      Intent's attributes
    :return:
    """

    app = Skill()

    return util.run_until_complete(
        app.test_intent(intent, translation, session=session, **kwargs)
    )



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Error middleware"""

import logging
import traceback
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError
from fastapi.encoders import jsonable_encoder

from skill_sdk.responses import ErrorCode

logger = logging.getLogger(__name__)

INTERNAL_ERROR = "Internal error"
BAD_REQUEST = "Bad request"
NOT_FOUND = "Not found"


def setup_middleware(app: FastAPI):
    """
    Setup error handlers (only those that have a corresponding error code in skill SPI):

        - 400: BAD_REQUEST
        - 404: NOT_FOUND
        - 422 is reraised as 400
        - 500: INTERNAL_ERROR

    :param app:
    :return:
    """

    @app.exception_handler(400)
    async def handle_400(request, exc):
        """Log the exception and return BAD_REQUEST"""

        logger.exception("%s %s: %s", request.method, request.url, repr(exc))
        return JSONResponse(
            status_code=400,
            content=dict(code=ErrorCode.BAD_REQUEST, text=BAD_REQUEST),
        )

    @app.exception_handler(404)
    async def handle_404(request, exc):
        """Log the exception and return NOT_FOUND"""

        logger.exception("%s %s: %s", request.method, request.url, repr(exc))
        return JSONResponse(
            status_code=404,
            content=dict(code=ErrorCode.NOT_FOUND, text=NOT_FOUND),
        )

    @app.exception_handler(RequestValidationError)
    async def handle_422(request, exc):
        """Unprocessable entity: raised by Pydantic validators"""

        logger.exception("%s %s: %s", request.method, request.url, repr(exc))
        return JSONResponse(
            status_code=422,
            content=dict(
                code=ErrorCode.BAD_REQUEST,
                text=BAD_REQUEST,
                detail=jsonable_encoder(exc.errors()),
            ),
        )

    @app.exception_handler(500)
    async def handle_500(request, exc):
        """Log the exception and return INTERNAL_ERROR"""

        logger.exception("500 error raised, returning internal error: %s", repr(exc))
        content = dict(
            code=ErrorCode.INTERNAL_ERROR,
            text=INTERNAL_ERROR,
        )

        # Output the exception in "debug" mode
        if request.app.debug:
            content.update(detail=jsonable_encoder(traceback.format_exc()))

        return JSONResponse(
            status_code=500,
            content=content,
        )



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill Designer UI"""

import time
import asyncio
import inspect
from json import dumps
import logging.handlers
from pathlib import Path
from typing import Any, Callable, Dict, List, Text, Tuple
import datetime

from jinja2 import Environment, FileSystemLoader
from fastapi import responses, FastAPI, Request, WebSocket, WebSocketDisconnect
from starlette.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles
from pydantic import validator

from skill_sdk.utils.util import CamelModel
from skill_sdk.intents.entities import AttributeV2, TimeRange, TimeSet

logger = logging.getLogger()

UI_ROOT = Path(__file__).absolute().parent
DEFAULT_MODULE = "impl"
DESIGNER = "Designer endpoints"
TEMPLATES = "templates"

#
#   Sample values for intent handlers
#

SAMPLE_BOOL = True
SAMPLE_INT = 42
SAMPLE_FLOAT = 42.0242
SAMPLE_COMPLEX = complex(42.02, 0.42)
SAMPLE_STRING = "string value"
SAMPLE_ATTR_V2: AttributeV2 = AttributeV2(id=1, value=SAMPLE_STRING)
SAMPLE_TIME_RANGE = TimeRange("2002-02-02T12:42:02Z/2022-02-02T00:12:42Z")
SAMPLE_TIME_SET = TimeSet("(XXXX-WXX-1,XXXX-WXX-5,P4D)")


def samples(type_=None):
    """
    Return a sample value for particular type (if specified), or list of supported types otherwise

    :param type_:
    :return:
    """
    _ = {
        "bool": SAMPLE_BOOL,
        "int": SAMPLE_INT,
        "float": SAMPLE_FLOAT,
        "complex": SAMPLE_COMPLEX,
        "str": SAMPLE_STRING,
        "timedelta": datetime.timedelta(),
        "datetime": datetime.datetime.now(),
        "date": datetime.date.today(),
        "time": datetime.datetime.now().time(),
        "TimeRange": SAMPLE_TIME_RANGE,
        "TimeSet": SAMPLE_TIME_SET,
        "AttributeV2": SAMPLE_ATTR_V2,
        "typing.List[str]": [SAMPLE_STRING],
    }
    return _.get(type_, None) if type_ is not None else list(_)


class Parameter(CamelModel):
    """Parameter record format"""

    # Parameter name
    name: Text

    # Parameter type
    type: Text

    # Required if non-empty and has no default
    required: bool = False

    # Sample value
    sample: Any

    # Values array
    values: List[Any] = []

    @validator("values", pre=True, always=True)
    def default(cls, v):  # pylint: disable=E0213
        return [v] if not isinstance(v, (list, tuple, type(None))) else (v or [])

    @property
    def as_code(self) -> Text:
        """
        Return string representation as Python parameter with type annotation

        :return:
        """
        value = f" = {repr(self.values[0])}" if self.values else None
        default = "" if self.required else " = None"
        return f"{self.name}: {self.type}{value or default}".replace("'", '"')

    @staticmethod
    def from_signature(name: Text, type_: Text, default: Any = None) -> "Parameter":
        """
        Create an instance from function signature

        :param name:
        :param type_:
        :param default:
        :return:
        """
        return Parameter(
            name=name,
            type=type_,
            required=default == inspect.Parameter.empty,
            sample=samples(type_),
            values=[default] if default not in (None, inspect.Parameter.empty) else [],
        )


class Intent(CamelModel):
    """Intent record format"""

    # Intent name
    name: Text

    # Handler (implementation) name as tuple (<module>, <function>)
    implementation: Tuple = ()

    @property
    def module(self):
        return self.implementation[0]

    @property
    def function(self):
        return self.implementation[1].lower()

    @validator("implementation", pre=True, always=True)
    def generate_name(cls, v, *, values):  # pylint: disable=E0213
        return v or (DEFAULT_MODULE, f"handle_{values['name']}")

    # List of intent handler parameters (entities)
    parameters: List[Parameter] = []

    @staticmethod
    def from_callable(name: Text, func: Callable) -> "Intent":
        """
        Create an instance from intent handler

        :param name:    intent name
        :param func:    handler function
        :return:
        """

        params = inspect.signature(func).parameters
        parameters = [
            Parameter.from_signature(
                p_name,
                getattr(param.annotation, "__name__", repr(param.annotation)),
                param.default,
            )
            for p_name, param in params.items()
        ]

        return Intent(
            name=name,
            implementation=(getattr(func, "__module__", DEFAULT_MODULE), func.__name__),
            parameters=parameters or [],
        )


def _signature_changed(intent: Intent, intents: Dict[Text, Callable]) -> bool:
    """
    Denotes if function signature has been altered in the UI

    :param intent:  an intent from UI
    :param intents: dictionary of intent implementations
    :return:
    """
    return (
        intent.json() != Intent.from_callable(intent.name, intents[intent.name]).json()
    )


async def get_types():
    """
    List supported attribute types
    """

    return responses.JSONResponse(content=[_ for _ in samples()])


async def get_intents(request: Request):
    """
    List intents and entity samples
    """

    return [
        Intent.from_callable(name, func) for name, func in request.app.intents.items()
    ]


class Notifier:
    """Notifies subscribed clients"""

    def __init__(self):
        self.connections: List[WebSocket] = []
        self.generator = self.get_notification_generator()

    async def get_notification_generator(self):
        while True:
            message = yield
            await self._notify(message)

    async def push(self, msg: Text):
        try:
            await self.generator.asend(msg)
        except RuntimeError:
            pass

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.connections.append(websocket)

    def remove(self, websocket: WebSocket):
        self.connections.remove(websocket)

    async def _notify(self, message: Text):
        living_connections = []
        while len(self.connections) > 0:
            # Looping is necessary in case a disconnection is handled
            # during await websocket.send_text(message)
            websocket = self.connections.pop()
            await websocket.send_text(message)
            living_connections.append(websocket)
        self.connections = living_connections

    async def worker(self, queue):
        """
        Async worker: waits for a log record and sends to subscribed clients

        :param queue:
        :return:
        """
        while True:
            record: logging.LogRecord = await queue.get()
            await self.push(dumps(record.__dict__))


notifier = Notifier()


async def ws_endpoint(websocket: WebSocket):
    """
    Dump logs output to websocket

    :param websocket:
    :return:
    """
    await notifier.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        try:
            notifier.remove(websocket)
        except ValueError:
            pass  # Shutting down (?)


def log_changes(app_intents: Dict[Text, Callable], intents: List[Intent]) -> bool:
    """
    Log differences between the running skill intents and a list received from the UI

    :param app_intents: running skill's intents (request.app.intents)
    :param intents:     intents list from UI editor
    :return:            True if there are changes in intents
    """

    # List of new intents from UI
    to_add = [_ for _ in intents if _.name not in app_intents]

    # List of intents missing in UI request (deleted)
    to_delete = [
        Intent.from_callable(name, func)
        for name, func in app_intents.items()
        if name not in [_.name for _ in intents]
    ]

    # List of intents those signatures are different from original
    to_update = [
        intent
        for intent in intents
        if intent.name in app_intents and _signature_changed(intent, app_intents)
    ]

    if len(to_add):
        logger.debug("Will create %s intent handler(s): %s", len(to_add), repr(to_add))

    if len(to_delete):
        logger.debug(
            "Will delete %s intent handler(s): %s", len(to_delete), repr(to_delete)
        )

    if len(to_update):
        logger.debug(
            "Will modify %s intent handler(s): %s", len(to_update), repr(to_update)
        )

    return any((to_add, to_delete, to_update))


def render_impl(intents: List[Intent]) -> Text:
    """
    Render Python code for intent handlers implementation

    :param intents:
    :return:
    """

    imports = []

    # Collect datetime related parameter types
    dt_types = [
        _.type
        for intent in intents
        for _ in intent.parameters
        if _.type in ("date", "time", "datetime", "timedelta")
    ]
    if dt_types:
        imports.append(f"from datetime import {', '.join(dt_types)}")

    # Collect SDK entities types
    sdk_types = [
        _.type
        for intent in intents
        for _ in intent.parameters
        if _.type in ("AttributeV2", "TimeRange", "TimeSet")
    ]
    if sdk_types:
        imports.append(f"from skill_sdk.entities import {', '.join(sdk_types)}")

    imports.append("from skill_sdk import skill, Response")

    return (
        Environment(loader=FileSystemLoader(UI_ROOT / TEMPLATES))
        .get_template("impl.j2")
        .render(imports=imports, intents=intents)
    )


def render_tests(intents: List[Intent]) -> Text:
    """
    Render unit tests code

    :param intents:
    :return:
    """

    return (
        Environment(loader=FileSystemLoader(UI_ROOT / TEMPLATES))
        .get_template("tests.j2")
        .render(intents=intents)
    )


def render_runner(intents: List[Intent]) -> Text:
    """
    Render runner ("app.py")

    :param intents:
    :return:
    """

    return (
        Environment(loader=FileSystemLoader(UI_ROOT / TEMPLATES))
        .get_template("app.j2")
        .render(intents=intents)
    )


def save(content: Text, path: Path, suffix: Text) -> Path:
    """
    Save the content to the file specified by path parameter,
    backup previously existing file

    :param content: file content
    :param path:    file name
    :param suffix:  backup file suffix
    :return:        absolute file path
    """
    if path.exists():
        backup = path.with_suffix(f".{suffix}")
        logger.info("File %s exists, backing up to %s", path, backup)
        backup.write_text(path.read_text())

    path.write_text(content)

    logger.info("Dump content to %s", path.absolute())
    return path.absolute()


async def post_intents(request: Request, intents: List[Intent]):
    """
    Generator: main entry point
    """

    if not log_changes(request.app.intents, intents):
        return responses.JSONResponse(
            content=dict(message="No changes"), status_code=204
        )

    # Render Python modules from templates
    suffix = str(time.time())
    impl = save(render_impl(intents), Path(DEFAULT_MODULE) / "__init__.py", suffix)
    tests = save(render_tests(intents), Path(DEFAULT_MODULE) / "test_impl.py", suffix)
    runner = save(render_runner(intents), Path("app.py"), suffix)

    # Reload intent handlers
    request.app.reload("app")

    return responses.JSONResponse(
        content=dict(impl=str(impl), tests=str(tests), runner=str(runner))
    )


def setup(app: FastAPI):
    """
    Install skill designer UI routes and CORS middleware

    :param app:
    :return:
    """

    # CORS middleware to allow requests from dev and prod UI
    app.add_middleware(
        CORSMiddleware,
        allow_origins=[
            "http://localhost:8080",
            "http://127.0.0.1:8080",
            "http://localhost:4242",
            "http://127.0.0.1:4242",
        ],
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # List types route
    app.add_api_route(
        "/types",
        get_types,
        name="List types",
        tags=[DESIGNER],
    )

    # List intents route
    app.add_api_route(
        "/intents",
        get_intents,
        name="List intents",
        tags=[DESIGNER],
    )

    # Generator route
    app.add_api_route(
        "/intents",
        post_intents,
        methods=["POST"],
        name="Generate intent handlers",
        tags=[DESIGNER],
    )

    # Websocket logs push
    app.add_websocket_route(
        "/logs",
        ws_endpoint,
        name="Skill Logs",
    )

    @app.on_event("startup")
    async def startup():
        """Initializes websocket listener"""

        # Prime the push notification generator
        try:
            await notifier.generator.asend(None)
        except StopAsyncIteration:
            pass

        # Silence the websocket.protocol logger
        logging.getLogger("websockets").setLevel(logging.CRITICAL)

        # Setup a queue that passes log records to websockets
        queue = asyncio.Queue()

        # Default format of the log record in front-end
        formatter = logging.Formatter("%(levelname)-8s %(name)s - %(message)s")

        handler = logging.handlers.QueueHandler(queue)
        handler.setLevel(logger.level)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        # Start the notifier's worker
        asyncio.ensure_future(notifier.worker(queue))

    # Delete root redirect
    app.router.routes = [
        route for route in app.router.routes if getattr(route, "path", None) != "/"
    ]

    # Mount root UI
    app.mount("/", StaticFiles(directory=str(UI_ROOT), html=True), name="Skill UI")



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Tracing middleware"""

import logging
from fastapi import FastAPI
from contextlib import ContextDecorator

logger = logging.getLogger(__name__)

try:
    from starlette_opentracing import StarletteTracingMiddleWare
    from opentracing import Tracer, global_tracer, set_global_tracer
except ModuleNotFoundError:
    logger.error(
        '"StarletteTracingMiddleWare" not found. Extra package is not installed. '
        'Make sure to install "skill-sdk[all]" to enable distributed tracing features.'
    )
    raise


class start_span(ContextDecorator):
    """
    Tracing helper/span wrapper.

    Can be used as both context manager and decorator:

        As context manager:
        >>> with start_span('span'):
        >>>    ...

        As decorator:
        >>> @start_span('span')
        >>> def decorated():
        >>>    ...

    """

    def __init__(self, operation_name, *args, **kwargs):
        self.operation_name = operation_name
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        tracer = global_tracer()
        self._span = tracer.start_span(self.operation_name, *self.args, **self.kwargs)
        return self._span.__enter__()

    def __exit__(self, _exc_type, _exc_value, _exc_traceback):
        return self._span.__exit__(_exc_type, _exc_value, _exc_traceback)


def setup(app: FastAPI, tracer: Tracer) -> None:
    """
    Setup tracing: register a middleware and set global tracer instance

    :param app:
    :param tracer:
    :return:
    """
    set_global_tracer(tracer)
    app.add_middleware(StarletteTracingMiddleWare, tracer=tracer)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Logging"""

import os
import re
import time
import json
import logging.config
from traceback import format_exc
from typing import Dict, Optional

from skill_sdk import config


def setup_logging(
    log_level: Optional[int] = None, log_format: Optional[config.FormatType] = None
):
    """
    Set log level either to the level provided as argument
    or get it from LOB_LEVEL environment var

    :param log_level:
    :param log_format:
    :return:
    """

    log_level = log_level or config.settings.LOG_LEVEL
    log_format = log_format or config.settings.LOG_FORMAT

    logging.getLogger().setLevel(log_level)
    logging.basicConfig(level=log_level)

    try:
        logging.config.dictConfig(get_config_dict(log_level, log_format))
    except KeyError:
        raise RuntimeError("Invalid log format: %s", repr(log_format))

    # Patch Uvicorn logger default formats and level
    if isinstance(log_level, str):
        log_level = logging.getLevelName(log_level)
    for _ in ("uvicorn", "uvicorn.error", "uvicorn.access", "uvicorn.asgi"):
        logger = logging.getLogger(_)
        #
        # Change loggers level to target,
        # unless the target level is DEBUG and Uvicorn loggers level is TRACE:
        # we don't have TRACE logging level
        #
        if not (log_level == logging.DEBUG and logger.level < logging.DEBUG):
            logger.setLevel(log_level)
        if log_format == config.FormatType.GELF:
            for handler in logger.handlers:
                handler.setFormatter(CloudGELFFormatter())


def tracing_headers() -> Dict:
    """Extract tracing headers from Starlette's context"""
    from skill_sdk.middleware import HeaderKeys, context

    try:
        return {
            key: context.data[key]
            for key in HeaderKeys
            if context.data[key] is not None
        }
    # If we're not inside a request
    except RuntimeError:
        return {}


class CloudGELFFormatter(logging.Formatter):
    """Graylog Extended Format (GELF) formatter"""

    def format(self, record: logging.LogRecord):
        from skill_sdk.middleware import HeaderKeys

        headers = tracing_headers()

        # Cloud log record format
        line = {
            # Timestamp in milliseconds
            "@timestamp": int(round(time.time() * 1000)),
            # Log message level
            "level": record.levelname,
            # Process id
            "process": os.getpid(),
            # Thread id
            "thread": str(record.thread),
            # Logger name
            "logger": record.name,
            # Log message
            "message": record.getMessage(),
            # Trace id
            "traceId": headers.get(HeaderKeys.trace_id, None),
            # Span id
            "spanId": headers.get(HeaderKeys.span_id, None),
            # Testing flag
            "testing": str(headers.get(HeaderKeys.testing_flag, False)).lower()
            in ("true", "1"),
            # Tenant: a skill is not aware of tenant
            "tenant": headers.get(HeaderKeys.tenant_id, None),
            # Magenta Transaction Id
            "magentaTransactionId": headers.get(
                HeaderKeys.magenta_transaction_id, None
            ),
        }

        if record.exc_info:
            line["_traceback"] = format_exc()

        return json.dumps(line)


def get_config_dict(log_level: int, log_format: config.FormatType) -> Dict:
    """Logging configuration dictionary"""

    conf = {
        "gelf": {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": {"class": "skill_sdk.log.CloudGELFFormatter"},
            },
            "handlers": {
                "default": {
                    "level": log_level,
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                },
            },
            "loggers": {
                "": {"handlers": ["default"], "level": log_level, "propagate": True},
            },
        },
        "human": {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "standard": {
                    "format": "%(asctime)s %(levelname)-8s %(name)s - %(message)s",
                }
            },
            "handlers": {
                "default": {
                    "level": log_level,
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                },
            },
            "loggers": {
                "": {"handlers": ["default"], "level": log_level, "propagate": True},
            },
        },
    }

    return conf[log_format]


def patch_logger():
    """
    Patch the `logging.Logger.isEnabledFor` method.

    :return:
    """

    from skill_sdk.middleware import HeaderKeys, context

    _super = logging.Logger.isEnabledFor

    def is_enabled_for(instance: logging.Logger, level):
        """
        Return True if "X-User-Debug-Log" flag is set

        :param instance:    logging.Logger instance
        :param level:       logging level
        :return:
        """
        try:
            return context.data[HeaderKeys.user_debug_log] or _super(instance, level)
        except (KeyError, RuntimeError):
            return _super(instance, level)

    logging.Logger.isEnabledFor = is_enabled_for


patch_logger()


###############################################################################
#                                                                             #
#   Optional formatter for GunicornLogger                                     #
#                                                                             #
###############################################################################
try:
    from gunicorn.glogging import Logger

    class GunicornLogger(Logger):
        """Gunicorn logger that formats log messages with CloudGELFFormatter"""

        def setup(self, cfg):
            self.loglevel = config.settings.LOG_LEVEL
            self.error_log.setLevel(self.loglevel)
            self.access_log.setLevel(self.loglevel)

            if config.settings.LOG_FORMAT == config.FormatType.GELF:
                self._set_handler(self.error_log, cfg.errorlog, CloudGELFFormatter())
                self._set_handler(self.access_log, cfg.errorlog, CloudGELFFormatter())

except ModuleNotFoundError:  # pragma: no cover
    pass


###############################################################################
#                                                                             #
#  Limit log message size, and hide tokens from logs                          #
#                                                                             #
###############################################################################
JWT_REGEX = re.compile(r"^eyJ[A-Za-z0-9_-]*\.[A-Za-z0-9._-]*")


def _trim(s):
    """Trim long string to LOG_ENTRY_MAX_STRING(+3) length"""

    return (
        s
        if not isinstance(s, str) or len(s) < config.settings.LOG_ENTRY_MAX_STRING
        else s[: config.settings.LOG_ENTRY_MAX_STRING] + "..."
    )


def _copy(d, hide_tokens: bool = False):
    """
    Recursively copy dictionary values, trimming long strings, and hiding possible JWT tokens
    """

    if isinstance(d, dict):
        return {k: _copy(v) for k, v in d.items()}
    elif isinstance(d, (list, tuple)):
        return [_copy(v) for v in d]
    else:
        return _trim(JWT_REGEX.sub("eyJ*****", d) if hide_tokens else d)


def prepare_for_logging(record, hide_tokens: bool = False):
    """
    Trim long strings before logging a record

    :param record:      value to log
    :param hide_tokens: specifies if JWT-like values should be replaced by asterisks
    :return:
    """
    return _copy(record, hide_tokens)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill invoke data response"""

from typing import Dict, Optional, Text
from skill_sdk.utils.util import CamelModel
from skill_sdk.responses.task import DelayedClientTask


class Result(CamelModel):
    """Skill result in machine-readable format"""

    class Config:
        """This model instances are mutable for backward compatibility"""

        allow_mutation = True

    # Result data
    data: Dict

    # True if passed over to the local client
    local: bool = True

    # Id of the controlled device
    target_device_id: Optional[Text]

    # Delayed client task
    delayed_client_task: Optional[DelayedClientTask]

    def __init__(
        self,
        data: Dict,
        *,
        local: bool = True,
        target_device_id: Text = None,
        delayed_client_task: DelayedClientTask = None,
    ) -> None:
        """
        Rewrite positional arguments to keyword ones for backward compatibility

        :param data:
        :param local:
        :param target_device_id:
        :param delayed_client_task:
        """
        super().__init__(
            data=data,
            local=local,
            target_device_id=target_device_id,
            delayed_client_task=delayed_client_task,
        )

    def __getitem__(self, *args):
        return self.data.__getitem__(*args)

    def __bool__(self):
        return any((self.data, self.target_device_id, self.delayed_client_task))

    def update(self, *args, **kwargs):
        """
        Update `data`

        :return:
        """
        return self.data.update(*args, **kwargs)

    def with_task(self, task: DelayedClientTask):
        """
        Add task to result

        :param task:
        :return:
        """
        return self.copy(update=dict(delayed_client_task=task))



#!/usr/bin/env python

"""Your application instance"""

from skill_sdk import skill
from impl import handler

# Create a skill instance
app = skill.init_app()

# Add "SMALLTALK__GREETINGS" intent handler
app.include(handler=handler)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Error responses"""

from enum import IntEnum
from typing import Any, Optional, Text
from pydantic import BaseModel


class ErrorCode(IntEnum):
    """Skill SPI error response codes"""

    # skill invoke with a wrong intent
    NOT_FOUND = 1

    # given third party token is invalid
    INVALID_TOKEN = 2

    # expected arguments like locale are missing
    BAD_REQUEST = 3

    # timeout
    TIMEOUT = 4

    # not considered situation, unhandled exception
    INTERNAL_ERROR = 999


class ErrorResponse(BaseModel):
    """
    An error response

    It can be returned explicitly from the intent handler or will be returned if calling the intent handler fails.

    The following combinations are defined:

    **wrong intent**
      ``{"code": 1, "text": "intent not found"}`` HTTP code: *404*

    **invalid token**
      ``{"code": 2, "text": "invalid token"}`` HTTP code: *400*

    **version, locale,… missing**
      ``{"code": 3, "text": "Bad request"}`` HTTP code: *400*

    **time out**
      ``{"code": 4, "text": "Time out"}`` HTTP code: *504*

    **unhandled exception**
      ``{"code": 999, "text": "internal error"}`` HTTP code: *500*

    """

    code_map = {1: (404,), 2: (400,), 3: (400,), 4: (504,), 999: (500,)}

    # Error code
    code: ErrorCode

    # Error message
    text: Text

    # Details (if in debug mode)
    detail: Optional[Any]



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill responses"""

from typing import Text

from skill_sdk.responses.info import SkillInfoResponse

from skill_sdk.responses.error import (
    ErrorCode,
    ErrorResponse,
)

from skill_sdk.responses.card import (
    Card,
    CardAction,
    ListItem,
    ListSection,
)

from skill_sdk.responses.command import (
    Command,
    AudioPlayer,
    Calendar,
    System,
    Timer,
)

from skill_sdk.responses.result import Result

from skill_sdk.responses.response import (
    SkillInvokeResponse,
    SkillInvokeResponse as Response,
    ResponseType,
    SessionResponse,
    _enrich,
)

from skill_sdk.responses.reprompt import Reprompt

from skill_sdk.responses.task import ClientTask, DelayedClientTask


def tell(text: Text, **kwargs) -> Response:
    """
    Wrapper to return Response of TELL type

    :param text:
    :param kwargs:
    :return:
    """
    return Response(**{**kwargs, **dict(text=text, type=ResponseType.TELL)})


def ask(text: Text, **kwargs) -> Response:
    """
    Wrapper to return Response of ASK type

    :param text:
    :param kwargs:
    :return:
    """
    return Response(**{**kwargs, **dict(text=text, type=ResponseType.ASK)})


def ask_freetext(text: Text, **kwargs) -> Response:
    """
    Wrapper to return Response of ASK_FREETEXT type

    :param text:
    :param kwargs:
    :return:
    """
    return Response(**{**kwargs, **dict(text=text, type=ResponseType.ASK_FREETEXT)})



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: utility functions"""

import sys
import argparse
import logging
import importlib
import pathlib
from typing import Optional, Text, Tuple
from types import ModuleType

from skill_sdk import config, skill, Skill


#
# Default module (folder) for handler implementations:
#
#   if folder is missing, it will be created if skill is in "development" mode
#   (ran with `vs develop`)
#
DEFAULT_MODULE = "impl"


def add_logging_options(parser: argparse.ArgumentParser) -> None:
    """
    Add common logging argument parameters

    :param parser:
    :return:
    """
    logging_arguments = parser.add_argument_group("Logging options")

    logging_arguments.add_argument(
        "-v",
        "--verbose",
        help="Set logging to INFO.",
        action="store_const",
        dest="loglevel",
        const=logging.INFO,
    )
    logging_arguments.add_argument(
        "-vv",
        "--debug",
        help="Set logging to DEBUG.",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
    )
    logging_arguments.add_argument(
        "-q",
        "--quiet",
        help="Set logging to ERROR.",
        action="store_const",
        dest="loglevel",
        const=logging.ERROR,
    )


def add_module_argument(
    parser: argparse.ArgumentParser,
    default: Optional[Text] = DEFAULT_MODULE,
) -> None:
    parser.add_argument("module", help="Run module", nargs="?", default=default)


def add_env_file_argument(
    parser: argparse.ArgumentParser,
    default: Optional[Text] = config.DOTENV_FILE,
) -> None:
    """
    Add env-file parameter argument:

    :param parser:
    :param default:
    :return:
    """
    parser.add_argument(
        "-e",
        "--env-file",
        default=default,
        help="Location of dotenv file.",
    )


def process_env_file(arguments):
    """
    Set .dotenv file location

    :param arguments:
    :return:
    """

    env_file = getattr(arguments, "env_file", None)
    if env_file is not None:
        config.Settings.Config.env_file = getattr(arguments, "env_file")


def import_module_app(import_from: Text) -> Tuple[ModuleType, Skill]:
    """
    Import application from python module or package

    :param import_from: module or package name, can be in following formats:
                        "app.py", "app:app" or "app_dir"

    :return:
    """
    module_str, _, app_str = import_from.partition(":")

    path = pathlib.Path(module_str)

    # Insert current directory to sys.path
    cwd = pathlib.Path("").absolute().__str__()
    if cwd not in sys.path:
        sys.path.insert(0, cwd)

    if path.suffix == ".py":
        module = importlib.import_module(path.stem)
    elif path.is_dir():
        module = importlib.import_module(module_str)
        for _ in path.iterdir():
            if _.is_file() and _.suffix == ".py" and _.stem != "__init__":
                importlib.import_module("." + _.stem, _.parent.name)
            elif _.is_dir() and "__init__.py" in [f.name for f in _.iterdir()]:
                importlib.import_module("." + _.stem, _.parent.name)
    else:
        module = importlib.import_module(module_str)

    # Extract the name of application variable, if specified (`main:app`)
    app = getattr(module, app_str, None)
    # Create a default app, if not
    if app is None:
        app = skill.init_app()

    # Add link to the originating module
    # for source reloading and intent handler updates from UI
    if app.debug:
        setattr(app, "_module", module)

    # Return imported module and the application instance
    return module, app



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors /
# copyright owners license this file to you under the MIT
# License (the "License"); you may not use this file
# except in compliance with the License.
# You may obtain a copy of the License at
#
# https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""Magenta Voice Skill SDK: version info"""

__name__ = "skill-sdk"
__description__ = "Magenta Voice Skill SDK for Python"
__url__ = "https://github.com/telekom/voice-skill-sdk/"
__version__ = "1.4.0"
__author__ = "Deutsche Telekom Voicification Suite"
__author_email__ = "vineet.hingorani@telekom.de"
__license__ = "MIT"
__copyright__ = "Copyright 2021, Deutsche Telekom AG"

__spi_version__ = "1.5.1"



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Logging middleware: extract headers for logging"""

from enum import Enum
from starlette_context.plugins.base import Plugin


class HeaderKeys(str, Enum):
    """Internal testing headers"""

    trace_id = "X-B3-TraceId"
    span_id = "X-B3-SpanId"
    tenant_id = "X-Tenant-Id"
    testing_flag = "X-Testing"
    user_debug_log = "X-User-Debug-Log"
    magenta_transaction_id = "Baggage-X-Magenta-Transaction-Id"


class TraceIdPlugin(Plugin):
    """Extracts trace ID from request"""

    key = HeaderKeys.trace_id


class SpanIdIdPlugin(Plugin):
    """Extracts span ID from request"""

    key = HeaderKeys.span_id


class TenantIdIdPlugin(Plugin):
    """**IMPORTANT**: for logging purpose only, do not base tenant-specific logic on this header

    Extracts tenant ID from request
    """

    key = HeaderKeys.tenant_id


class TestingFlagPlugin(Plugin):
    """
    Extracts "testing" flag from request:
    testing flag distinguishes testing traffic from production
    """

    key = HeaderKeys.testing_flag


class UserDebugLogPlugin(Plugin):
    """
    Extracts "user debug log" flag from request:
    the flag is set to temporary activate debug logging for the user
    """

    key = HeaderKeys.user_debug_log


class MagentaTransactionId(Plugin):
    """
    Extracts Magenta transaction ID
    """

    key = HeaderKeys.magenta_transaction_id



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Prometheus middleware"""

import time
import logging
from functools import partial, wraps
from typing import Any, Callable, Text
from contextlib import contextmanager, ContextDecorator
from fastapi import FastAPI
import httpx

from skill_sdk.config import settings

logger = logging.getLogger(__name__)

HTTP_REQUESTS_LATENCY_SECONDS = "http_requests_latency_seconds"
HTTP_PARTNER_REQUEST_COUNT = "http_partner_request_count"

try:
    from starlette_exporter import PrometheusMiddleware, handle_metrics
    from prometheus_client import Counter, Histogram
except ModuleNotFoundError:
    logger.error(
        '"PrometheusMiddleware" not found. Extra package is not installed. '
        'Make sure to install "skill-sdk[all]" to export Prometheus metrics.'
    )
    raise


class Prometheus(PrometheusMiddleware):
    """Middleware factory"""

    @staticmethod
    def requests_latency():
        """HTTP requests latency histogram"""

        metric_name = HTTP_REQUESTS_LATENCY_SECONDS
        if metric_name not in PrometheusMiddleware._metrics:
            PrometheusMiddleware._metrics[metric_name] = Histogram(
                metric_name,
                "HTTP Requests Latency in second",
                ("job", "version", "operation"),
            )
        return PrometheusMiddleware._metrics[metric_name]

    @staticmethod
    def partner_requests_count():
        """HTTP requests to partner services counter"""

        metric_name = HTTP_PARTNER_REQUEST_COUNT
        if metric_name not in PrometheusMiddleware._metrics:
            PrometheusMiddleware._metrics[metric_name] = Counter(
                metric_name,
                "HTTP Requests for services",
                ("job", "partner_name", "status"),
            )
        return PrometheusMiddleware._metrics[metric_name]


class prometheus_latency(ContextDecorator):  # noqa
    """
    Prometheus latency wrapper. Can be used as context manager and decorator:

        >>> # As context manager:
        >>> with prometheus_latency("operation_name"):
        >>>     ...

        >>> # As decorator:
        >>> @prometheus_latency("operation_name")
        >>> def decorated():
        >>>     ...

    """

    def __init__(self, operation_name: Text):
        self.operation_name = operation_name

    def __enter__(self):
        self.begin = time.perf_counter()

    def __exit__(self, exc_type, exc, exc_tb):
        end = time.perf_counter()

        labels = [settings.SKILL_NAME, settings.SKILL_VERSION, self.operation_name]
        Prometheus.requests_latency().labels(*labels).observe(end - self.begin)


def _inc_partner_call(partner_name: Text, response: httpx.Response) -> None:
    try:
        status_code = response.status_code
        labels = [settings.SKILL_NAME, partner_name, status_code]
        Prometheus.partner_requests_count().labels(*labels).inc()
    except AttributeError as ex:
        logger.error("Cannot log a partner call: %s", repr(ex))


@contextmanager
def partner_call(partner_name: Text, func: Callable[..., Any]):
    """
    Context manager to count HTTP requests to partner service

        >>> with partner_call('partner-service', client.get) as get:
        >>>     response = get(URL)

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        """
        The actual logic: collect status code from wrapped function call and increase a metric counter

        :param args:
        :param kwargs:
        :return:
        """
        response = func(*args, **kwargs)
        _inc_partner_call(partner_name, response)

        return response

    yield wrapper


def count_partner_calls(partner_name: Text) -> Callable[[httpx.Response], None]:
    """
    Response hook to count HTTP requests to partner service,
    can be attached to `skill_sdk.requests.Client`/`AsyncClient`

        >>> from skill_sdk.requests import AsyncClient
        >>> from skill_sdk.middleware.prometheus import count_partner_calls
        >>>
        >>> async with AsyncClient(
        >>>     response_hook=count_partner_calls('partner-service')
        >>> ) as client:
        >>>     response = await client.get(URL)

    """

    return partial(_inc_partner_call, partner_name)


def setup(app: FastAPI) -> None:
    """
    Setup the Prometheus exporter middleware,
    and an app route to export the metrics

    :param app:
    :return:
    """

    app.add_middleware(
        Prometheus,
        app_name=settings.SKILL_NAME,
        prefix="http",
    )

    route = getattr(settings, "PROMETHEUS_ENDPOINT", "/prometheus")
    app.add_route(route, handle_metrics)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Type hints processing and intent handler invoke"""

import inspect
import logging
from typing import (
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Text,
    Type,
    Union,
)

from pydantic.utils import lenient_issubclass

from skill_sdk.utils.util import run_in_executor
from skill_sdk.intents import entities
from skill_sdk.intents import Context, Request, Session, RequestContextVar
from skill_sdk.responses import Response, _enrich, ErrorResponse

from functools import wraps, reduce, partial


logger = logging.getLogger(__name__)
AnyType = Type[Any]
AnyFunc = Callable[..., Any]
ErrorHandlerType = Callable[[Text, "EntityValueException"], Union[Awaitable, Response]]


class EntityValueException(Exception):
    """
    Entity values conversion exception
    """

    def __init__(self, ex, value=None, func=None, *args):
        self.__cause__ = ex
        self.value = value
        self.func = func
        super().__init__(*args)


async def invoke(handler: AnyFunc, request: Request) -> Union[Response, ErrorResponse]:
    """
    Invoke intent handler:

        awaits the call if handler is a coroutine
        runs in executor if handler is `def`

    :param handler:
    :param request:
    :return:
    """

    with RequestContextVar(request=request):
        logger.debug(
            "Calling intent %s with handler: %s",
            repr(request.context.intent),
            repr(handler),
        )

        if inspect.iscoroutinefunction(handler):
            response = await handler(request)
        else:
            response = await run_in_executor(handler, request)

        if not isinstance(response, ErrorResponse):
            response = _enrich(response)

        logger.debug("Intent call result: %s", repr(response))
        return response


def _is_subtype(cls: Any, class_or_tuple: Any) -> bool:
    """
    Return true if class is a generic subclass of class_or_tuple.

    :param cls:
    :param class_or_tuple:   a class or tuple of class objects
    :return:
    """
    return bool(
        getattr(cls, "__origin__", None)
        and lenient_issubclass(cls.__origin__, class_or_tuple)
    )


def _is_attribute_v2(annotation) -> bool:
    """
    Return true if annotation contains AttributeV2

    :param annotation:
    :return:
    """
    if isinstance(annotation, (list, tuple)):
        return _is_attribute_v2(next(iter(annotation), None))
    if _is_subtype(annotation, (List, Tuple)):
        args = getattr(annotation, "__args__", None) or annotation
        return _is_attribute_v2(next(iter(args), None))
    return _is_subtype(annotation, entities.AttributeV2) or lenient_issubclass(
        annotation, entities.AttributeV2
    )


def list_functor(annotation):
    """Convert to List of type values"""

    to_type = next(iter(annotation), None)
    if _is_subtype(to_type, entities.AttributeV2):
        return partial(map, attr_v2_functor(to_type.__args__)), list
    return partial(map, entities.converter(to_type)), list


def attr_v2_functor(annotation):
    """Convert to AttributeV2 with type value"""

    to_type = next(iter(annotation), None)
    return partial(entities.AttributeV2, mapping=entities.converter(to_type))


def _converters(annotation) -> Tuple:
    """Construct converter functions"""

    if isinstance(annotation, (list, tuple)):
        converter = list_functor(annotation)
    elif _is_subtype(annotation, (List, Tuple)):
        converter = list_functor(annotation.__args__)
    elif _is_subtype(annotation, entities.AttributeV2):
        converter = entities.get_entity, attr_v2_functor(annotation.__args__)
    elif lenient_issubclass(annotation, (Request, Context, Session)):
        converter = (lambda c: c,)  # No-op
    else:
        # Get a single attribute and convert to type
        converter = entities.get_entity, entities.converter(annotation)
    return converter


def get_converters(
    func_name: Text,
    parameters: AbstractSet[Tuple[Text, inspect.Parameter]],
    reduce_func: Callable,
) -> Dict[Text, partial]:
    """
    Helper: Constructs converter functions
        a dict of {"name": (f1, f2, ...)} where f1, f2, ... will be applied to handler arguments

    :param func_name:   function name (used just to throw the exception)
    :param parameters:  function parameters
    :param reduce_func: final reduce function
    :return:
    """
    converters = {}
    for name, param in list(parameters):
        if param.annotation == inspect.Parameter.empty:
            raise ValueError(
                f"Function {func_name} - parameter '{name}' has no type annotation"
            )

        converters[name] = partial(reduce_func, _converters(param.annotation))

    return converters


def apply(value, func: Callable[[Text], Any]) -> Optional[Any]:
    """
    Apply callable to value, returning EntityValueException if conversion error occurs

    :param value:
    :param func:
    :return:
    """
    if value is None:
        return None

    try:
        return func(value)
    except Exception as ex:  # NOSONAR
        logger.exception(
            "Exception converting %s with %s: %s ", repr(value), repr(func), repr(ex)
        )
        return EntityValueException(ex, value=value, func=func)


def is_handler(f: AnyFunc) -> bool:
    """Returns `True` if function is an intent handler"""
    return getattr(f, "__intent_handler__", False)


def get_inner(f: AnyFunc) -> AnyFunc:
    """Skip inner @intent_handler decorators"""

    try:
        return get_inner(getattr(f, "__wrapped__")) if is_handler(f) else f
    except AttributeError:
        return f


def intent_handler(
    func: AnyFunc = None,
    silent: bool = True,
    error_handler: ErrorHandlerType = None,
):
    """
    Generic intent handler decorator:
    Will check the handler parameters and supply entity values according to the type hints

        an intent handler function is supposed to have the following signature:
            handler(request: skill_sdk.Request, entity_one: typing.Any, entity_two: typing.Any, *)

        to receive a date in a handler function, use:
            @intent_handler
            handler(date: datetime.date)

        to receive an array of integer values, use:
            @intent_handler
            handler(int_list: [int])

        to suppress the conversion errors, should you want to handle exceptions, set `silent` to `True`.
        The decorator will return exception as value:
            @intent_handler(silent=True)

    :param func:    decorated function (can be `None` if decorator used without call)
    :param silent:  if `True`, an exception occurred during conversion will not be raised and returned as value
    :param error_handler:  if set, will be called if conversion error occurs, instead of a decorated function
    :return:
    """
    if isinstance(func, bool):
        silent, func = func, None

    def handler_decorator(_func: AnyFunc) -> AnyFunc:
        """The entry point to the decorator"""

        inner = get_inner(_func)
        signature = inspect.signature(inner)
        parameters = signature.parameters.items()
        converters = get_converters(inner.__name__, parameters, partial(reduce, apply))

        target = partial(  # type: ignore
            _parse_and_call,
            func=inner,
            converters=converters,
            silent=silent,
            error_handler=error_handler,
        )

        if inspect.iscoroutinefunction(inner):

            if error_handler and not inspect.iscoroutinefunction(error_handler):
                raise ValueError(
                    f"Error handler {repr(error_handler.__name__)} must be a coroutine"
                )

            @wraps(inner)
            async def wrapper(*args, **kwargs) -> Response:
                """Entry point to async intent handler"""

                # if the first arg is not a request, set request to None
                if not (args and isinstance(args[0], Request)):
                    args = (None, *args)

                return await target(*args, **kwargs)

        else:

            @wraps(inner)
            def wrapper(*args, **kwargs) -> Response:
                """Entry point to sync handler"""

                # if the first arg is not a request, set request to None
                if not (args and isinstance(args[0], Request)):
                    args = (None, *args)

                return target(*args, **kwargs)

        setattr(wrapper, "__intent_handler__", True)
        return wrapper

    return handler_decorator(func) if func else handler_decorator


def _parse_and_call(
    request: Request,
    *args,
    func: AnyFunc,
    converters: Dict[Text, partial],
    silent: bool,
    error_handler: ErrorHandlerType = None,
    **kwargs,
):
    """Extract parameters from invoke request and call intent handler"""

    signature = inspect.signature(func)
    parameters = signature.parameters.items()

    if isinstance(request, Request):
        # Proceed with skill invoke request if first parameter is invoke request
        kw = _parse_request(request, parameters)
        logger.debug("Collected arguments: %s", repr(kw))

        ba = signature.bind(**kw)
        arguments = {
            name: converters[name](value) for name, value in ba.arguments.items()
        }

        logger.debug("Converted arguments to: %s", repr(arguments))
        ba.arguments.update(arguments)

        # raises EntityValueException if not silent mode
        errors = _parse_errors(arguments, silent)

        return (
            _log_and_call("Normal call", func, *ba.args, **ba.kwargs)
            if not (errors and error_handler)
            else _log_and_call("Exception during conversion", error_handler, *errors)
        )

    else:
        # If we're called without invoke request as first argument, this is a direct call:
        #   we do not parse the context and simply pass arguments to the decorated function
        return _log_and_call(
            "Direct call",
            func,
            *args,
            **kwargs,
        )


def _log_and_call(message: Text, func: Callable, *args, **kwargs):
    """Helper to log debug message and call inner function"""

    logger.debug(
        "%s: calling %s with: %s, %s",
        message,
        repr(func.__name__),
        repr(args),
        repr(kwargs),
    )
    return func(*args, **kwargs)


def _parse_errors(arguments: Dict[Text, Any], silent: bool = True):
    """Extract type/value conversion exceptions"""
    errors = next(
        iter(
            (name, ex)
            for name, ex in arguments.items()
            if isinstance(ex, EntityValueException)
        ),
        (),
    )
    if errors and not silent:
        raise errors[1]
    return errors


def _as_attributes(attrs_v2: List[entities.AttributeV2]) -> List[Any]:
    """
    Convert entities: list of AttributesV2 to simple values list

    :param attrs_v2:
    :return:
    """
    return [_.value for _ in attrs_v2]


def _parse_request(
    request: Request, parameters: AbstractSet[Tuple[Text, inspect.Parameter]]
) -> Dict[Text, Any]:
    """
    Parse attributes from skill invoke request

    :param request:
    :param parameters:
    :return:
    """
    result = {
        # original request
        name: request if lenient_issubclass(param.annotation, Request)
        # if annotated as Context, use request.context
        else request.context if lenient_issubclass(param.annotation, Context)
        # if annotated as Session, use request.session
        else request.session if lenient_issubclass(param.annotation, Session)
        # look up attributesV2, if annotated as AttributeV2
        else request.context.attributes_v2.get(name)
        if _is_attribute_v2(param.annotation)
        # look up the standard attributes
        else request.context.attributes.get(name)
        for name, param in parameters
    }
    return result



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Internationalization"""

import re
import random
import logging
import subprocess
from pathlib import Path
from functools import reduce
from types import MappingProxyType
from typing import Dict, Iterable, List, Optional, Mapping, Text, Tuple, Union

import yaml
from yaml.reader import ReaderError
from yaml.scanner import ScannerError
from babel import dates, lists, support

# Place your `[lang].yaml` / `[lang].po` files to `locale` directory
LOCALE_DIR = "locale"

# Place combined YAML translations to `locales.yaml` file
LOCALE_FILE = "locales.yaml"

PROGRAM = "pybabel"
PROGRAM_NOT_FOUND = f'Failed to launch "{PROGRAM} %s": not found. Make sure "{PROGRAM}" is in your PATH.'

RE_TRANSLATIONS = re.compile(r"^[a-z]{2}(-[A-Z]{2})?$")

logger = logging.getLogger(__name__)


def get_locale_dir(locale_dir: Text = None) -> Path:
    """Returns locales folder location"""
    return Path(locale_dir or LOCALE_DIR)


def make_lazy(func, alt=None):
    """
    Make lazy translation function

    :param func:    function to call
    :param alt:     alternative function if translation is not set
    :return:
    """

    def lazy_func(*args, **kwargs):
        """Lazy translations wrapper"""

        from skill_sdk.intents import r

        try:
            return getattr(r.get_translation(), func)(*args, **kwargs)
        except TypeError:
            logger.error("Calling translation functions outside of request context.")
        except AttributeError as e:
            logger.exception("%s", repr(e))
        return alt(*args, **kwargs) if callable(alt) else None

    return lazy_func


_ = make_lazy("gettext", lambda m, *a, **kw: m)
_n = make_lazy(
    "ngettext", lambda singular, plural, n, *a, **kw: singular if n == 1 else plural
)
_a = make_lazy("getalltexts", lambda m, *a, **kw: [m])


class TranslationError(Exception):
    """
    Exception raised when a translation could not be performed due to a missing ``.mo`` file, a missing translation
    key or if there are no suitable translations available in the text service.
    """


MESSAGE_KEY_DEFAULT_SEPARATOR = " "


class Message(str):
    """String object that encapsulates formatting parameters"""

    # Message id
    key: Text

    # Message string (un-formatted)
    value: Text

    # Positional arguments
    args: Tuple

    # Keyword arguments
    kwargs: Dict

    def __new__(cls, value, key=None, *args, **kwargs):
        """
        Create a message with msgstr/msgid and format parameters

        :return:
        """
        message = (
            value.format(*args, **kwargs)
            if isinstance(value, str) and (args or kwargs)
            else value
        )
        string = super().__new__(cls, message)
        string.key = key or value
        string.args = args
        string.kwargs = kwargs
        string.value = value
        return string

    def format(self, *args, **kwargs) -> "Message":
        """
        Create and return new Message object with given format parameters

        :return:
        """
        message = Message(self.value, self.key, *args, **kwargs)
        return message

    def __add__(self, other: Union["Message", Text]) -> "Message":
        """
        Concatenate messages (or Message and str)

        :param other:
        :return:
        """
        if isinstance(other, Message):
            value = self.value + other.value
            key = MESSAGE_KEY_DEFAULT_SEPARATOR.join((self.key, other.key)).strip()
            args = self.args + other.args
            kwargs = {**self.kwargs, **other.kwargs}
        else:
            value = self.value + other
            key = (
                # Enclose string in quotes for readability
                MESSAGE_KEY_DEFAULT_SEPARATOR.join((self.key, f'"{other}"')).strip()
                if other
                else self.key
            )
            args, kwargs = self.args, self.kwargs

        return Message(value, key, *args, **kwargs)

    def join(self, iterable: Iterable[Union["Message", Text]]):
        """
        Join messages in iterable and return a concatenated Message.

        :param iterable:
        :return:
        """
        return reduce(lambda x, y: x + self + y, iterable)

    def strip(self, __chars: Optional[Text] = None) -> "Message":
        """
        Return new Message object with stripped value

        :return:
        """
        message = Message(
            self.value.strip(__chars), self.key, *self.args, **self.kwargs
        )
        return message


class Translations(support.Translations):
    """Lazy translations, return Message object instead of formatted string"""

    def __init__(self, lang: Text = None, f=None):
        self.lang = lang
        try:
            with f.open(mode="rb") as fp:
                super().__init__(fp)
        except AttributeError:
            super().__init__()

    def gettext(self, message, *args, **kwargs) -> Message:
        return Message(super().gettext(message), message, *args, **kwargs)

    def ngettext(self, singular, plural, n, *args, **kwargs) -> Message:
        return Message(super().ngettext(singular, plural, n), singular, *args, **kwargs)

    def format_list(self, elements: List[Text], style="standard"):
        """
        Join list elements
            [items, item2, item3] -> 'item1, item2 and item3'

        :param elements:
        :param style:
        :return:
        """
        return lists.format_list(elements, style=style, locale=self.lang)

    # Backward compatibility
    nl_join = format_list

    def nl_build(self, header: Text, elements: List[Text]) -> Text:
        """
        Build list in natural language:
            (header, [items, item2, item3]) -> 'Header: item1, item2, and item3.'

        :param header:      list header
        :param elements:    list elements
        :return:
        """
        return f'{Message(": ").join((header, self.format_list(elements)))}.'

    def format_datetime(self, datetime=None, format="medium", tzinfo=None) -> Text:
        """Format datetime according to the locale"""
        return dates.format_datetime(datetime, format, tzinfo, self.lang)

    def format_date(self, date=None, format="medium") -> Text:
        """Format date according to the locale"""
        return dates.format_date(date, format, self.lang)

    def format_time(self, time=None, format="medium", tzinfo=None) -> Text:
        """Format time according to the locale"""
        return dates.format_time(time, format, tzinfo, self.lang)

    def format_timedelta(
        self,
        delta,
        granularity="second",
        threshold=0.85,
        add_direction=False,
        format="long",
    ) -> Text:
        """Format a time delta according to the rules of the given locale"""
        return dates.format_timedelta(
            delta, granularity, threshold, add_direction, format, self.lang
        )


class MultiStringTranslation(Translations):
    """Translations that allows single key to have multiple values"""

    def _parse(self, fp):
        """
        Load catalogue from YAML file

        :param fp:
        :return:
        """

        try:
            self._load_catalog(yaml.safe_load(fp))
        except (ReaderError, ScannerError) as ex:
            logger.exception(
                "Could not load translations from %s: %s", repr(fp), repr(ex)
            )
            raise RuntimeError from ex

    def _load_catalog(self, catalog):
        # Support Ruby-format translations with language code as top level key:
        pointer = catalog
        keys = list(catalog)
        if len(catalog) == 1 and RE_TRANSLATIONS.match(keys[0]):
            # Double-check if top level language code is the same as file name
            if self.lang == keys[0]:
                pointer = catalog[keys[0]]
            else:
                raise RuntimeError(
                    "Invalid language code %s when loading %s.",
                    repr(keys[0]),
                    repr(self.lang),
                )

        self._catalog = {
            k: v if isinstance(v, list) else [v] for k, v in pointer.items()
        }

    @staticmethod
    def from_dict(lang: Text, catalog: Dict) -> "MultiStringTranslation":
        translation = MultiStringTranslation(lang)
        translation._load_catalog(catalog)
        return translation

    def __repr__(self):
        return f"<{type(self).__name__}: {repr(self.lang)}>"

    def gettext(self, message, *args, **kwargs):
        logger.debug("Translating message %s to %s", repr(message), repr(self.lang))
        try:
            candidates = self._catalog[message]
            logger.debug("%s candidates: %s", len(candidates), repr(candidates))
            return Message(random.choice(candidates), message, *args, **kwargs)
        except LookupError:
            logger.warning("No translation for key: %s", repr(message))
            return super().gettext(message, *args, **kwargs)

    def ngettext(self, singular, plural, n, *args, **kwargs):
        logger.debug("Translating %s/%s/%s to %s", singular, plural, n, self.lang)
        return self.gettext(singular if n == 1 else plural, *args, **kwargs)

    def getalltexts(self, key, *args, **kwargs):
        logger.debug("Retrieving all translation messages for %s in %s", key, self.lang)
        try:
            candidates = self._catalog[key]
            logger.debug("%s candidates: %s", len(candidates), repr(candidates))
            return [Message(value, key, *args, **kwargs) for value in candidates]
        except LookupError:
            logger.warning("No translation for key: %s", key)
            return [super().gettext(key, *args, **kwargs)]


def compile_locales(locale_dir: Text = None, force: bool = False):
    """
    Compile all languages available in locale_dir:
    launches `pybabel compile` to compile .po to .mo files

    :param locale_dir:
    :param force:       force compilation even if *.mo files exist
    :return:
    """
    command = "compile"

    for po_file in get_locale_dir(locale_dir).glob("*.po"):

        mo_file = po_file.with_suffix(".mo")
        if mo_file.exists() and not force:
            logger.info("Skipping %s: %s exists", po_file.name, mo_file)
            continue

        logger.info("Compiling %s ...", po_file.name)
        try:

            result = subprocess.check_output(
                [
                    PROGRAM,
                    command,
                    "-i",
                    str(po_file),
                    "-o",
                    str(mo_file),
                ],
                text=True,
                stderr=subprocess.STDOUT,
            )
            logger.info(result)

        except FileNotFoundError:
            logger.error(PROGRAM_NOT_FOUND, command)

        except subprocess.CalledProcessError as ex:
            logger.error("Failed to compile %s: %s", po_file.name, ex.stdout)
            raise


def _load_all(locale_file: Text = LOCALE_FILE) -> Dict[Text, MultiStringTranslation]:
    """
    Load translations from a single locale file

    :param locale_file:
    :return:
    """
    try:
        with Path(locale_file).open() as fp:
            catalog = yaml.safe_load(fp)
            if not all((RE_TRANSLATIONS.match(x) for x in catalog)):
                raise RuntimeError(
                    "One or more invalid language codes found: %s.", repr(list(catalog))
                )

            return {
                lang: MultiStringTranslation.from_dict(lang, catalog[lang])
                for lang in catalog
            }

    except FileNotFoundError:
        logger.debug("%s not found.", locale_file)
        return {}


def _load_yaml(locale_dir: Text = None) -> Dict[Text, MultiStringTranslation]:
    """
    Load multi-string translations from YAML files

    :param locale_dir:
    :return:
    """

    logger.info("Loading YAML translations...")

    return {
        yaml_file.stem: MultiStringTranslation(yaml_file.stem, yaml_file)
        for yaml_file in get_locale_dir(locale_dir).glob("*.yaml")
        if RE_TRANSLATIONS.match(yaml_file.stem)
    }


def _load_gettext(locale_dir: Text = None) -> Dict[Text, Translations]:
    """
    Load `gettext` translations from *.po/*.mo files

    :param locale_dir:
    :return:
    """

    logger.info("Loading gettext translations...")

    compile_locales(locale_dir)
    return {
        mo_file.stem: Translations(mo_file.stem, mo_file)
        for mo_file in get_locale_dir(locale_dir).glob("*.mo")
        if RE_TRANSLATIONS.match(mo_file.stem)
    }


def load_translations(
    locale_file: Text = LOCALE_FILE, locale_dir: Text = LOCALE_DIR
) -> Mapping[Text, Translations]:
    """
    Load local languages from locale_file or locale_dir

    :param locale_file: YAML file with combined translations  ("locales.yaml")
    :param locale_dir:  Folder to load YAML/PO/MO files with single translation per file ("locale/")
    :return:
    """

    translations = (
        _load_all(locale_file) or _load_yaml(locale_dir) or _load_gettext(locale_dir)
    )

    if translations:
        logger.info("Loaded: %s", list(translations))
    else:
        logger.info("No local translations found.")

    return MappingProxyType(translations)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""Geolocation service"""

import logging
from typing import Dict, List, Optional, Text

import httpx

from skill_sdk.utils.util import CamelModel, root_validator
from skill_sdk.services.base import BaseService

logger = logging.getLogger(__name__)


#
#   The models below reflect location service's response structure
#


class GeoLocation(CamelModel):
    """Geo-location: latitude/longitude"""

    lat: float
    lng: float


class AddressComponents(CamelModel):
    """Address components: city and postal code"""

    city: Text
    postal_code: Optional[Text]


class Address(CamelModel):
    """Address: country with address components"""

    country: Text
    address_components: AddressComponents


class Location(CamelModel):
    """Location consists of an address and time-zone"""

    address: Address
    time_zone: Text = "Europe/Berlin"


class FullLocation(GeoLocation, Location):
    """Full location consists of an address component and geographical coordinates"""

    ...


class FullAddress(CamelModel):
    """Complete address including geo-coordinates"""

    lat: Optional[float]
    lon: Optional[float]  # ATTN: "lon" not "lng" (!)

    country: Optional[Text]
    city: Optional[Text]
    postal_code: Optional[Text]
    street_name: Optional[Text]
    street_number: Optional[Text]

    def __bool__(self):
        return any(
            (
                self.country,
                self.city,
                self.postal_code,
                self.street_name,
                self.street_number,
            )
        )


class FullAddressList(CamelModel):
    """Result of address lookup service: list of FullAddress"""

    __root__: List[Optional[FullAddress]]

    def __bool__(self):
        return bool(self.__root__)


class AddressLookupQuery(CamelModel):
    """Query to address lookup service"""

    country: Optional[Text]
    postalcode: Optional[Text]  # noqa
    street_name: Optional[Text]
    street_number: Optional[Text]

    lang: Optional[Text]

    # Max number of search results
    limit: int

    @root_validator(pre=True)
    def query_required(cls, values: Dict):  # pylint: disable=E0213
        if not any(
            (
                values.get("country"),
                values.get("postalcode"),
                values.get("street_name"),
                values.get("street_number"),
            )
        ):
            raise ValueError("Query is missing or empty!")
        return values


class GeoLookupQuery(CamelModel):
    """Request to location service"""

    country: Optional[Text]
    city: Optional[Text]
    postalcode: Optional[Text]  # noqa

    lang: Optional[Text]

    @root_validator(pre=True)
    def query_required(cls, values: Dict):  # pylint: disable=E0213
        if not any(
            (values.get("country"), values.get("city"), values.get("postalcode"))
        ):
            raise ValueError("Query is missing or empty!")
        return values


class LocationService(BaseService):
    """Location service with geo-coding"""

    VERSION = 1
    NAME = "location"

    async def forward_lookup(
        self,
        *,
        country: Text = None,
        city: Text = None,
        postalcode: Text = None,  # noqa
        lang: Text = None,
    ) -> FullLocation:
        """
        Forward lookup: resolve geo-coordinates from textual address components

        :param country:
        :param city:
        :param postalcode:
        :param lang:
        :return:
        """
        async with self.async_client as client:
            params = GeoLookupQuery(
                country=country, city=city, postalcode=postalcode, lang=lang
            )

            data = await client.get(f"{self.url}/geo", params=params.dict())
            return FullLocation(**data.json())

    async def reverse_lookup(self, lat: float, lng: float) -> Address:
        """
        Reverse lookup: resolve address components from geo-coordinates

        :param lat:
        :param lng:
        :return:
        """

        async with self.async_client as client:
            location = GeoLocation(lat=lat, lng=lng)

            data = await client.get(f"{self.url}/reversegeo", params=location.dict())
            return Address(**data.json())

    async def address_lookup(
        self,
        *,
        country: Text = None,
        postalcode: Text = None,  # noqa
        street_name: Text = None,
        street_number: Text = None,
        lang: Text = None,
        limit: int = 1,
    ) -> Optional[FullAddressList]:
        """
        Address lookup: get a list of addresses (with geo-coordinates) for a given query

        :param country:
        :param postalcode:
        :param street_name:
        :param street_number:
        :param lang:
        :param limit:
        :return:
        """
        async with self.async_client as client:
            params = AddressLookupQuery(
                country=country,
                postalcode=postalcode,
                street_name=street_name,
                street_number=street_number,
                lang=lang,
                limit=limit,
            )

            data = await client.request(
                "GET",
                f"{self.url}/address",
                params=params.dict(),
                exclude=[httpx.codes.NOT_FOUND],
            )
            return FullAddressList.parse_obj(data.json() if data.text else [])

    async def device_location(self) -> FullAddress:
        """
        Retrieves device location: whatever user has setup in cApp, plus geo coordinates

        (device serial number gets decoded from the service token,
         so in order to use this function, parent request must contain the service token)

        @return:
        """

        async with self.async_client as client:
            data = await client.request(
                "GET", f"{self.url}/device-location", exclude=[httpx.codes.NOT_FOUND]
            )
            return FullAddress(**data.json() if data.text else {})



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""CLI: "init" command"""

import sys
import argparse
import pkg_resources
from pathlib import Path
from distutils.dir_util import copy_tree


def execute(args: argparse.Namespace) -> None:
    """
    Initialize an empty skill project

    :param args:
    :return:
    """
    import questionary

    # Projects folder path
    path = Path(
        args.out
        if args.out is not None
        else questionary.path(
            "Enter a path to create your project [default: '.']", only_directories=True
        ).ask()
    ).resolve()

    if not path.is_dir():
        exit(f"{repr(path)} not found.")

    #
    # Intent handlers are by default in "impl" folder,
    # if the folder exists, we have to ask user's permission to overwrite it
    #
    if (path / "impl").exists():
        confirm = questionary.confirm(
            f'Project directory {repr(path / "impl")} exists. Overwrite?', default=True
        ).ask()
        if not confirm:
            sys.exit("Exiting...")

    scaffold_path = pkg_resources.resource_filename(__name__, "scaffold")
    copy_tree(scaffold_path.__str__(), path.__str__())
    print(f"Project initialized at {repr(path.absolute())}.")


def add_subparser(subparsers) -> None:
    """
    Command arguments parser

    :param subparsers:
    :return:
    """

    init_parser = subparsers.add_parser(
        "init",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Initialize a new project.",
        description="Initializes a project template, with sample 'SMALLTALK__GREETINGS' intent handler, "
        "configuration and unit tests are included.",
    )

    init_parser.add_argument(
        "--name",
        default="my-skill",
        help="Name of the skill.",
    )

    init_parser.add_argument(
        "--out",
        default=None,
        help="Directory to create the project.",
    )

    init_parser.set_defaults(command=execute)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Client commands"""

from enum import Enum
from typing import Dict, Optional, Text

#
# Note: client kits are Python:
#   we use original field names,
#   not camelCased aliases
from skill_sdk.utils.util import BaseModel


class KitType(Text, Enum):
    """
    Client kits available for cloud skills

    """

    AUDIO_PLAYER = "audio_player"
    CALENDAR = "calendar"
    SYSTEM = "system"
    TIMER = "timer"


class Kit(BaseModel):
    """
    Client kit

    """

    kit_name: KitType
    action: Text
    parameters: Optional[Dict]


class Command(BaseModel):
    """
    Generic client command

    """

    use_kit: Kit

    def __init__(self, kit_name, action, **kwargs):
        super().__init__(
            use_kit=Kit(kit_name=kit_name, action=action, parameters=kwargs or None)
        )


class AudioPlayer(Command):
    """
    Kit for handling audio player/radio functions

    """

    class Action(Text, Enum):
        """Available action commands"""

        PLAY_STREAM = "play_stream"
        PLAY_STREAM_BEFORE_TEXT = "play_stream_before_text"
        STOP = "stop"
        PAUSE = "pause"
        RESUME = "resume"

    class ContentType(Text, Enum):
        """Defined channel/content types"""

        RADIO = "radio"
        VOICEMAIL = "voicemail"

    def __init__(self, action, **kwargs):
        super().__init__(KitType.AUDIO_PLAYER, action, **kwargs)

    @staticmethod
    def play_stream(url: Text) -> "AudioPlayer":
        """
        Start playing a generic internet stream, specified by "url" parameter

        :param url:
        :return:
        """
        return AudioPlayer(AudioPlayer.Action.PLAY_STREAM, url=url)

    @staticmethod
    def play_stream_before_text(url: Text) -> "AudioPlayer":
        """
        Start playing a stream, before pronouncing the response

        :param url:
        :return:
        """
        return AudioPlayer(AudioPlayer.Action.PLAY_STREAM_BEFORE_TEXT, url=url)

    @staticmethod
    def stop(content_type: ContentType = None, text: Text = None) -> "AudioPlayer":
        """
        Stop currently playing media (voicemail, radio, content tts),
            optionally say text BEFORE stopping

        :param content_type:
        :param text:
        :return:
        """
        content_type = content_type or AudioPlayer.ContentType.RADIO
        text = text or ""

        return AudioPlayer(
            AudioPlayer.Action.STOP,
            content_type=content_type,
            notify=dict(not_playing=text),
        )

    @staticmethod
    def pause(content_type: ContentType = None, text: Text = None) -> "AudioPlayer":
        """
        Pause playback, optionally say text AFTER playback paused

        :param content_type:
        :param text:
        :return:
        """
        content_type = content_type or AudioPlayer.ContentType.RADIO
        text = text or ""

        return AudioPlayer(
            AudioPlayer.Action.PAUSE,
            content_type=content_type,
            notify=dict(not_playing=text),
        )

    @staticmethod
    def resume(content_type: ContentType = None) -> "AudioPlayer":
        """
        Resume paused media, say response text before resuming

        :param content_type:
        :return:
        """
        content_type = content_type or AudioPlayer.ContentType.RADIO

        return AudioPlayer(AudioPlayer.Action.RESUME, content_type=content_type)


class Calendar(Command):
    """
    Calendar kit: snooze calendar alarm, cancel snooze

    """

    class Action(Text, Enum):
        """Available action commands"""

        SNOOZE_START = "snooze_start"
        SNOOZE_CANCEL = "snooze_cancel"

    def __init__(self, action, **kwargs):
        super().__init__(KitType.CALENDAR, action, **kwargs)

    @staticmethod
    def snooze_start(snooze_seconds: int = None) -> "Calendar":
        """
        Snooze calendar alarm by optional number of seconds

        :param snooze_seconds:
        :return:
        """
        return Calendar(Calendar.Action.SNOOZE_START, snooze_seconds=snooze_seconds)

    @staticmethod
    def snooze_cancel() -> "Calendar":
        """
        Cancel current snooze

        :return:
        """
        return Calendar(Calendar.Action.SNOOZE_CANCEL)


class System(Command):
    """
    System functions kit

    """

    class Action(Text, Enum):
        """Available action commands"""

        STOP = "stop"
        PAUSE = "pause"
        RESUME = "resume"
        NEXT = "next"
        PREVIOUS = "previous"
        SAY_AGAIN = "say_again"
        VOLUME_UP = "volume_up"
        VOLUME_DOWN = "volume_down"
        VOLUME_TO = "volume_to"

    class SkillType(Text, Enum):
        """Defined channel/skill types"""

        TIMER = "Timer"
        CONVERSATION = "Conversation"
        MEDIA = "Media"

    def __init__(self, action, **kwargs):
        super().__init__(KitType.SYSTEM, action, **kwargs)

    @staticmethod
    def stop(skill_type: SkillType = None) -> "System":
        """
        Send a `Stop` event: stops a foreground activity on the device.
        If there was another activity in background, it will gain focus.

        :param skill_type:  Stop a skill-related activity, or everything, if no skill specified
        :return:
        """
        return System(System.Action.STOP, skill=skill_type)

    @staticmethod
    def pause() -> "System":
        """
        Pause currently active content (if supported)

        :return:
        """
        return System(System.Action.PAUSE)

    @staticmethod
    def resume() -> "System":
        """
        Resume media (if paused)

        :return:
        """
        return System(System.Action.RESUME)

    @staticmethod
    def next() -> "System":
        """
        Switch to next item in content channel

        :return:
        """
        return System(System.Action.NEXT)

    @staticmethod
    def previous() -> "System":
        """
        Switch to previous item in content channel

        :return:
        """
        return System(System.Action.PREVIOUS)

    @staticmethod
    def say_again() -> "System":
        """
        Repeat last uttered sentence (from the dialog channel)

        :return:
        """
        return System(System.Action.SAY_AGAIN)

    @staticmethod
    def volume_up() -> "System":
        """
        Increase the volume one notch

        :return:
        """
        return System(System.Action.VOLUME_UP)

    @staticmethod
    def volume_down() -> "System":
        """
        Decrease the volume one notch

        :return:
        """
        return System(System.Action.VOLUME_DOWN)

    @staticmethod
    def volume_to(value: int = 0) -> "System":
        """
        Set the volume to an absolute value (0-10)

        :return:
        """
        if value not in range(10):
            raise ValueError("Value %s in not in expected range (0 - 10)", repr(value))

        return System(System.Action.VOLUME_TO, value=value)


class Timer(Command):
    """
    Timer kit: start/stop Timer/Reminder/Alarm animation

        After animation started, it keeps running for max. 10 min if no user interaction happens

    """

    class Action(Text, Enum):
        """Available action commands"""

        SET_TIMER = "set_timer"
        CANCEL_TIMER = "cancel_timer"

    def __init__(self, action, **kwargs):
        super().__init__(KitType.TIMER, action, **kwargs)

    @staticmethod
    def set_timer():
        """
        Fire up a "timer" animation

        :return:
        """
        return Timer(Timer.Action.SET_TIMER)

    @staticmethod
    def cancel_timer():
        """
        Cancel currently running "timer"

        :return:
        """
        return Timer(Timer.Action.CANCEL_TIMER)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""Base for internal services"""

import logging
from os import environ
from typing import Dict, Text

from aiobreaker import CircuitBreaker

from skill_sdk.intents import r
from skill_sdk.requests import AsyncClient, Client

logger = logging.getLogger(__name__)

DEFAULT_AUTH_TOKEN = "cvi"

# Default timeout when accessing the service
DEFAULT_SERVICE_TIMEOUT = 10


class BaseService:
    """
    The base for cloud services

    """

    # Service version
    VERSION: int = 0

    # Service name
    NAME: Text = "base"

    # Circuit breaker: shared between client instances
    circuit_breaker: CircuitBreaker

    # Base service URL
    url: Text = ""

    # Timeout value
    timeout: float = DEFAULT_SERVICE_TIMEOUT

    #
    # Specifies if authorization header should be added to the request
    #
    add_auth_header: bool = True

    def __init__(
        self,
        url: Text,
        *,
        internal: bool = True,
        timeout: float = DEFAULT_SERVICE_TIMEOUT,
        headers: Dict[Text, Text] = None,
        add_auth_header: bool = None,
        auth_token: Text = DEFAULT_AUTH_TOKEN,
    ) -> None:
        self.url = url
        self.internal = internal
        self.timeout = timeout
        self.auth_token = auth_token
        self._headers = headers or {}

        # `True` for internal services, `False` for external, unless specified explicitly
        self.add_auth_header = (
            add_auth_header if add_auth_header is not None else internal
        )

        # circuit breaker with defaults: (fail_max=5, timeout_duration=60 sec)
        self.circuit_breaker = CircuitBreaker()

    def auth_header(self):
        """Add "Authorization" header for bearer auth

        :return:
        """

        # Get CVI token from current request
        try:
            token = r.context.tokens[self.auth_token]
            logger.debug("Adding auth token to authorization header.")
            return {"Authorization": f"Bearer {token}"}
        except (AttributeError, KeyError):
            logger.error(
                "No auth token %s found in the current request.", repr(self.auth_token)
            )
            return {}

    @property
    def headers(self) -> Dict:

        # Default request headers
        _headers: Dict[Text, Text] = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        # 'HOSTNAME' is the environment value provided by k8s (actually the pod name)
        user_agent = environ.get("HOSTNAME") or None
        if user_agent is not None:
            _headers.update({"User-Agent": user_agent})

        # If inside a request, add "Content-Language"
        if r and r.context:
            _headers.update({"Content-Language": r.context.locale})

        # Add "Authorization" header if requested
        if self.add_auth_header:
            _headers.update(self.auth_header())

        _headers.update(self._headers)

        logger.debug("Client headers: %s", repr(_headers))
        return _headers

    @property
    def client(self) -> Client:
        """Creates new client with circuit breaker"""

        return Client(
            internal=self.internal,
            base_url=self.url,
            headers=self.headers,
            timeout=self.timeout,
            circuit_breaker=self.circuit_breaker,
        )

    @property
    def async_client(self) -> AsyncClient:
        """Creates new async client with circuit breaker"""

        return AsyncClient(
            internal=self.internal,
            base_url=self.url,
            headers=self.headers,
            timeout=self.timeout,
            circuit_breaker=self.circuit_breaker,
        )



#
# voice-skill-sdk
#
# (C) 2021, YOUR_NAME (YOUR COMPANY)
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""Unit tests suite"""

import pytest
from skill_sdk import skill
from . import handler


@pytest.fixture
def app():
    """
    Pytest fixture to create an app for testing

    :return:
    """
    app = skill.init_app()
    app.include(handler=handler)
    return app


@pytest.mark.asyncio
async def test_intent_async(app):
    """
    This is a sample unit test of an intent in asynchronous context

    :param app:
    :return:
    """
    response = await app.test_intent("SMALLTALK__GREETINGS")
    assert response.text == "HELLOAPP_HELLO"


def test_intent_sync():
    """
    This is a sample unit test of an intent in sync context

    :return:
    """
    response = skill.test_intent("SMALLTALK__GREETINGS")
    assert response.text == "HELLOAPP_HELLO"



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Internationalization"""

import logging
import subprocess
from pathlib import Path
from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Text, Union

import yaml
from yaml.representer import Representer
from skill_sdk.i18n import PROGRAM, PROGRAM_NOT_FOUND, get_locale_dir

logger = logging.getLogger(__name__)


def extract_translations(
    modules: List[Text], locale_dir: Text = None
) -> Optional[Path]:
    """
    Extract translatable strings from Python modules and write translations to `messages.pot`

    :param modules: List of Python modules to scan
    :param locale_dir:
    :return:
    """
    command = "extract"
    path = get_locale_dir(locale_dir)
    if not path.exists():
        path.mkdir(parents=True)

    output = path / "messages.pot"
    try:
        result = subprocess.check_output(
            [
                PROGRAM,
                command,
                f"--input-dirs={','.join(modules)}",
                f"--output={str(output)}",
            ],
            text=True,
            stderr=subprocess.STDOUT,
        )
        logger.debug(result)

        logger.info("Translation template written to %s", repr(output))
        return output

    except FileNotFoundError:
        logger.error(PROGRAM_NOT_FOUND, command)

    except subprocess.CalledProcessError as ex:
        logger.error("Failed to extract translations: %s", repr(ex.stdout))

    return None


def init_locales(
    template: Path, locales: List[Text], locale_dir: Text = None, force: bool = False
) -> bool:
    """
    Create empty .po file in locale_dir

    :param template:    Template (.pot) file to create translation
    :param locales:     List of translations to initialize, eg. ['en', 'de', 'fr']
    :param locale_dir:  Locale folder
    :param force:       If `True`, try to unlink the [locale].po file first

    :return:            `True` if all locales have been initialized, `False` if error occurred
    """
    command = "init"
    for locale in locales:
        output = get_locale_dir(locale_dir) / f"{locale}.po"

        logger.info("Creating %s ...", repr(output))
        try:
            if output.exists() and force:
                output.unlink()

            subprocess.check_call(
                [
                    PROGRAM,
                    command,
                    f"--locale={locale}",
                    f"--input-file={template}",
                    f"--output-file={str(output)}",
                ]
            )

        except FileNotFoundError:
            logger.error(PROGRAM_NOT_FOUND, command)
            return False

        except subprocess.CalledProcessError as ex:
            logger.error("Failed to create %s: %s", repr(output), repr(ex))
            return False

    return True


def _translate(lines: Iterator, messages: Dict[Text, Union[Text, List[Text]]]) -> List:
    """
    Update lines from .po file with translated messages dict

    :param lines:
    :param messages:
    :return:
    """
    translated = []
    for line in lines:
        if line.strip().startswith("msgid"):
            translated.append(line)
            msgid = line.strip().split(" ")[-1].strip("'\"")
            msgstr = messages.get(msgid)
            if isinstance(msgstr, (list, tuple)):
                msgstr = next(iter(msgstr), None)
            if isinstance(msgstr, str):
                msgstr = msgstr.replace('"', '\\"')  # Escape double quotes
                msgstr = msgstr.strip()  # Strip blanks and new lines
                msgstr = msgstr.replace("\n", '"\n"')  # Add quotes to new lines
            try:
                line = f'msgstr "{msgstr}"' if msgstr else next(lines)
                next(lines)
            except StopIteration:
                pass
        translated.append(line)
    return translated


def translate_locale(
    locale: Text, messages: Dict[Text, Union[Text, List[Text]]], locale_dir: Text = None
) -> Optional[List]:
    """
    Read data from .po file and update it with translated messages

    :param locale:
    :param messages:
    :param locale_dir:
    :return:
    """
    po_file = get_locale_dir(locale_dir) / f"{locale}.po"
    try:
        logger.info("Translating %s ...", po_file.name)
        with po_file.open() as f:
            lines = iter(f.readlines())
            return _translate(lines, messages)
    except (AttributeError, KeyError, FileNotFoundError) as ex:
        logger.error("Failed to translate %s: %s", po_file.name, repr(ex))
        return None


def update_translation(
    locale: Text, messages: Dict[Text, Union[Text, List[Text]]], locale_dir: Text = None
) -> Optional[Path]:
    """
    Update .po file with translated messages

    :param locale:
    :param messages:
    :param locale_dir:
    :return:
    """
    po_file = get_locale_dir(locale_dir) / f"{locale}.po"
    translated = translate_locale(locale, messages, locale_dir)
    if translated:
        with po_file.open("w+") as f:
            logger.info("Updating %s ...", po_file.name)
            f.writelines(translated)
            return po_file
    else:
        logger.info("Nothing to translate in %s", po_file.name)
        return None


def _download_full_catalog(
    download_url: Text, scope: Text, token: Text = None, tenant: Text = None
) -> Dict:
    """
    Download a complete translation catalog from text service

    :param download_url:
    :param token:
    :param tenant:
    :return:
    """

    from skill_sdk.services.text import TextService

    logger.info(f"Downloading translations from {download_url}...")

    headers = {"X-Application-Authentication": f"Bearer {token}"} if token else {}
    headers.update({"X-Tenant": tenant}) if tenant else None

    service = TextService(download_url, scope, headers=headers, add_auth_header=False)
    return service.admin_get_full_catalog()


def download_translations(
    download_url: Text,
    scope: Text,
    token: Text = None,
    tenant: Text = None,
    force: bool = False,
):
    """
    Load translations from text service and save to locale/{language}.yaml

    :param download_url:    Text services URL to download translations
    :param scope:           Translation catalogue scope (skill name)
    :param token:           Bearer authentication token (X-Application-Authentication)
    :param tenant:          Tenant for authentication   (X-Tenant)
    :param force:           Overwrite existing translations if exist
    :return:
    """

    catalog = _download_full_catalog(download_url, scope, token, tenant)

    if catalog:

        yaml.add_representer(defaultdict, Representer.represent_dict)
        for locale in catalog:
            yaml_file = (get_locale_dir() / locale).with_suffix(".yaml")
            if yaml_file.exists() and not force:
                logger.error(
                    '"%s exists and no "--force" specified. Skipping...', yaml_file
                )
            else:
                logger.info("Saving %s to %s", repr(locale), yaml_file)
                with yaml_file.open("w+") as f:
                    yaml.dump(catalog[locale], f, allow_unicode=True)



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""HTTP sync/async clients with circuit breaker"""

from typing import Callable, Iterable, List, Union
import logging
from warnings import warn

import httpx
from httpx import codes, HTTPError, Response  # noqa

from aiobreaker import (
    CircuitBreaker,
    CircuitBreakerState,  # noqa
)

from skill_sdk.config import settings
from skill_sdk.log import tracing_headers

logger = logging.getLogger(__name__)

DEFAULT_REQUESTS_TIMEOUT = settings.REQUESTS_TIMEOUT


class Client(httpx.Client):
    """
    Sync HTTP client with a circuit breaker
    """

    def __init__(
        self,
        *,
        internal: bool = False,
        circuit_breaker: CircuitBreaker = None,
        timeout: Union[int, float] = None,
        exclude: Iterable[codes] = None,
        response_hook: Callable[[httpx.Response], None] = None,
        **kwargs,
    ) -> None:
        """
        Construct sync client

        :param internal:        identifies a request to an internal service
                                propagate the tracing headers, if the request is internal

        :param circuit_breaker: optional circuit breaker, DEFAULT_CIRCUIT_BREAKER if not set
        :param timeout:         optional timeout for a request
        :param exclude:         list of HTTP status codes that are treated as "normal" (no exception is raised)
        :param response_hook:   function to be executed after a response is received (with response as argument)
        :param kwargs:          keyword arguments passed over to request
        """
        self.internal = internal

        # If no custom circuit breaker supplied, we'll create a new instance
        self.circuit_breaker = circuit_breaker or CircuitBreaker()

        self.exclude = tuple(exclude) if exclude else ()
        super().__init__(timeout=timeout or DEFAULT_REQUESTS_TIMEOUT, **kwargs)
        if response_hook:
            self.event_hooks = dict(response=[response_hook])

    def request(
        self,
        *args,
        exclude: Iterable[codes] = None,
        **kwargs,
    ):
        exclude = exclude or self.exclude

        # Propagate tracing headers if request is created as "internal"
        if self.internal:
            logger.debug("Internal service, adding tracing headers.")
            kwargs["headers"] = {
                **(kwargs.get("headers", None) or {}),
                **tracing_headers(),
            }

        @self.circuit_breaker
        def _inner_call(*a, **kw):
            """Wraps Client.request"""
            _r = super(Client, self).request(*a, **kw)

            if _r.status_code in exclude:
                logger.debug("Status code %s is excluded", _r.status_code)
            else:
                _r.raise_for_status()

            return _r

        try:
            result = _inner_call(*args, **kwargs)
            logger.debug("HTTP completed with status code: %d", result.status_code)

        except HTTPError as e:
            logger.error(
                "HTTP request [%s, %s] failed with error: %s",
                repr(args),
                repr(kwargs),
                repr(e),
            )
            raise
        return result


class AsyncClient(httpx.AsyncClient):
    """
    Async client with a circuit breaker
    """

    def __init__(
        self,
        *,
        internal: bool = False,
        circuit_breaker: CircuitBreaker = None,
        timeout: Union[int, float] = None,
        exclude: List[codes] = None,
        response_hook: Callable[[httpx.Response], None] = None,
        **kwargs,
    ) -> None:
        """
        Construct sync client

        :param internal:        identifies a request to an internal service
                                propagate the tracing headers, if the request is internal

        :param circuit_breaker: optional circuit breaker, DEFAULT_CIRCUIT_BREAKER if not set
        :param timeout:         optional timeout for a request
        :param exclude:         list of HTTP status codes that are treated as "normal" (no exception is raised)
        :param response_hook:   function to be executed after a response is received (with response as argument)
        :param kwargs:          keyword arguments passed over to request
        """
        self.internal = internal

        # If no custom circuit breaker supplied, we'll create a new instance
        self.circuit_breaker = circuit_breaker or CircuitBreaker()

        self.exclude = tuple(exclude) if exclude else ()
        super().__init__(timeout=timeout or DEFAULT_REQUESTS_TIMEOUT, **kwargs)
        if response_hook:
            self.event_hooks = dict(response=[response_hook])

    async def request(
        self,
        *args,
        exclude: Iterable[codes] = None,
        **kwargs,
    ):
        exclude = exclude or self.exclude

        # Propagate tracing headers if request is created as "internal"
        if self.internal:
            logger.debug("Internal service, adding tracing headers.")
            kwargs["headers"] = {
                **(kwargs.get("headers", None) or {}),
                **tracing_headers(),
            }

        @self.circuit_breaker
        async def _inner_call(*a, **kw):
            """Wraps Client.request"""
            _r = await super(AsyncClient, self).request(*a, **kw)

            if _r.status_code in exclude:
                logger.debug("Status code %s is excluded", _r.status_code)
            else:
                _r.raise_for_status()

            return _r

        try:
            result = await _inner_call(*args, **kwargs)
            logger.debug("HTTP completed with status code: %d", result.status_code)

        except HTTPError as e:
            logger.error(
                "HTTP request [%s, %s] failed with error: %s",
                repr(args),
                repr(kwargs),
                repr(e),
            )
            raise
        return result


class CircuitBreakerSession(Client):
    """**DEPRECATED**: HTTP(s) session with a circuit breaker.

    Renamed to `skill_sdk.requests.Client`
    The name is left for backward compatibility

    """

    def __init__(self, *args, **kwargs) -> None:
        warn(
            '"requests.CircuitBreakerSession" is deprecated.\n'
            'Please use "requests.Client" if you need a circuit breaker or "httpx.Client" if not.',
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(*args, **kwargs)



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Skill invoke response"""

import copy
from enum import Enum
from typing import Any, Dict, Optional, List, Text, Union

from skill_sdk import i18n
from skill_sdk.i18n import Message
from skill_sdk.utils.util import CamelModel
from skill_sdk.responses.card import Card, ListSection
from skill_sdk.responses.command import Command
from skill_sdk.responses.result import Result
from skill_sdk.responses.task import ClientTask


class ResponseType(str, Enum):
    """
    Response types:

        TELL    pronounce the text on the device and end the session

        ASK     pronounce the text and wait for user's intent as response

        ASK_FREETEXT    pronounce the text and wait for user response as text

    """

    TELL = "TELL"
    ASK = "ASK"
    ASK_FREETEXT = "ASK_FREETEXT"


class PushNotification(CamelModel):
    """Messaging data to be sent to the client or app"""

    # The message
    message_payload: Optional[Text]

    # Target device name
    target_name: Optional[Text]


class SessionResponse(CamelModel):
    """Session attributes in the response"""

    attributes: Dict[Text, Text]


class SkillInvokeResponse(CamelModel):
    """Skill invoke response"""

    # Text to send to the device
    text: Union[Message, Text]

    # Response type
    type: ResponseType = ResponseType.TELL

    # Card to create
    card: Optional[Card]

    # Push notification to send
    push_notification: Optional[PushNotification]

    # Skill result data
    result: Optional[Result]

    # Device session
    session: Optional[SessionResponse]

    def __init__(
        self, text: Text, type_: ResponseType = None, result: Dict = None, **data: Any
    ) -> None:
        """
        Accept `text` and `type_` as positional arguments
            and `result` as dictionary for backward compatibility

        :param text:    text to pronounce
        :param type_:   response type
        :param data:
        """
        if "type" in data and type_ is not None:
            raise ValueError(
                f"Ambiguous response type: 'type_'={type_} and 'type='{data['type']}."
            )

        params: Dict[str, Any] = dict(text=text)
        if type_ is not None:
            params.update(type=type_)

        if result and isinstance(result, Dict):
            params.update(result=Result(result["data"] if "data" in result else result))

        super().__init__(**{**data, **params})

    def dict(self, *args, **kwargs) -> Dict[Text, Any]:
        """
        Dump the request into JSON suitable to be returned to the dialog manager.
        :param context: the context of the request
        """

        result = self.result or Result(dict())
        # Export string key and format parameters from Message object
        if isinstance(self.text, i18n.Message):
            result.update(
                key=self.text.key,
                value=self.text.value,
                args=self.text.args,
                kwargs=self.text.kwargs,
            )

        # Required properties
        resp: Dict[Text, Any] = dict()
        resp.update(type=self.type, text=self.text)

        # Optional properties
        if self.card:
            resp.update(card=self.card.dict())
        if result:
            resp.update(result=result.dict())
        if self.push_notification:
            resp.update(pushNotification=self.push_notification)
        if self.session:
            resp.update(session={"attributes": self.session.attributes})

        return resp

    def with_card(
        self,
        card: Card = None,
        *,
        icon_url: Text = None,
        title_text: Text = None,
        type_description: Text = None,
        image_url: Text = None,
        prominent_text: Text = None,
        action_prominent_text: Text = None,
        text: Text = None,
        sub_text: Text = None,
        media_url: Text = None,
        list_sections: List[ListSection] = None,
    ) -> "SkillInvokeResponse":
        """
        Attach Card to a response

        :param card:
        :param icon_url:
        :param title_text:
        :param type_description:
        :param image_url:
        :param prominent_text:
        :param action_prominent_text:
        :param text:
        :param sub_text:
        :param media_url:
        :param list_sections:
        :return:
        """
        return self.copy(
            update=dict(
                card=card
                or Card(
                    icon_url=icon_url,
                    title_text=title_text,
                    type_description=type_description,
                    image_url=image_url,
                    prominent_text=prominent_text,
                    action_prominent_text=action_prominent_text,
                    text=text,
                    sub_text=sub_text,
                    media_url=media_url,
                    list_sections=list_sections,
                )
            )
        )

    def with_command(self, command: Command) -> "SkillInvokeResponse":
        """
        Add a command to execute on the client

        :param command:
        :return:
        """
        result = Result(data=command.dict())
        return self.copy(update=dict(result=result))

    def with_notification(
        self,
        message: PushNotification = None,
        message_payload: Text = None,
        target_name: Text = None,
    ) -> "SkillInvokeResponse":
        """
        Add a command to execute on the client

        :param message:
        :param message_payload:
        :param target_name:
        :return:
        """
        message = message or PushNotification(
            message_payload=message_payload, target_name=target_name
        )
        return self.copy(update=dict(push_notification=message))

    def with_session(self, **attributes) -> "SkillInvokeResponse":
        """
        Add attributes (key -> value) to keep in session storage

            (valid only for ResponseType.ASK/ASK_FREETEXT:
             ResponseType.TELL immediately ends the session)

        :param attributes:
        :return:
        """

        if self.type == ResponseType.TELL:
            raise ValueError(f"Response type: {self.type} ends the session.")

        session = SessionResponse(attributes=attributes)
        return self.copy(update=dict(session=session))

    def with_task(self, task: ClientTask):
        """
        Add a delayed client task

        :param task:
        :return:
        """
        result = self.result or Result(data={})
        return self.copy(update=dict(result=result.with_task(task)))


def _enrich(response: SkillInvokeResponse) -> SkillInvokeResponse:
    """
    Post-process a skill invoke response:

        - If simple text is returned - convert to Response
        - If session attributes are present - add to response

    :param response:
    :return:
    """
    from skill_sdk.intents.request import r

    if isinstance(response, str):
        # Convert string response to Response object
        response = SkillInvokeResponse(text=response)

    #
    # Copy session attributes from global request,
    # unless response is TELL, that ends the session
    #
    if response.type != ResponseType.TELL and r.session.attributes:
        attributes = copy.deepcopy(r.session.attributes)
        return response.with_session(**attributes)

    return response



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Intent entities and conversion functions"""

import re
import logging
import datetime
import functools
from typing import Any, Callable, Dict, Generic, List, Optional, Text, TypeVar, Union
from dateutil import parser, rrule
from dateutil.tz import tzutc, gettz
import isodate
from pydantic import root_validator

from skill_sdk.utils.util import CamelModel


logger = logging.getLogger(__name__)
T = TypeVar("T")


def get_entity(entities: List[Any]) -> Optional[Any]:
    """Silently return first element of a list, or unchanged value otherwise"""
    return (
        next(iter(entities), None) if isinstance(entities, (list, tuple)) else entities
    )


@functools.singledispatch
def to_datetime(value) -> datetime.datetime:
    """Parse datetime string"""
    return parser.parse(value)


@functools.singledispatch
def to_date(value) -> datetime.date:
    """Parse datetime and return date component"""
    return to_datetime(value).date()


@functools.singledispatch
def to_time(value) -> datetime.time:
    """Parse datetime and return time component"""
    return to_datetime(value).time()


@functools.singledispatch
def to_timedelta(value) -> datetime.timedelta:
    """Parse duration string and return timedelta"""
    return isodate.parse_duration(value)


@to_timedelta.register(datetime.timedelta)
def __to_timedelta(value: datetime.timedelta):
    """No-op if already converted"""
    return value


def on_off_to_boolean(value: str) -> bool:
    """
    Converts ON_OFF entity value to boolean
    The value might be any string of ``0``, ``1``, ``on``, ``off``, ``true``, ``false``, case insensitive.

    :param value: the entity value
    """
    logger.debug("converting value %s to bool", repr(value))
    if not isinstance(value, str):
        raise ValueError(f"The value {value.__repr__()} is not a string.")
    if value.lower() in ("on", "true", "yes", "1"):
        return True
    if value.lower() in ("off", "false", "no", "0"):
        return False
    raise ValueError(f"{value} is not a proper on/off value.")


class TimeRange:
    """
    Date/Time range representation
    """

    __slots__ = ("begin", "end")

    begin: Optional[datetime.datetime]
    end: Optional[datetime.datetime]

    def __init__(self, value: str):
        self.begin, self.end = [
            parser.parse(v) if v else None for v in value.split("/")
        ]

    def __contains__(
        self, value: Union[datetime.datetime, datetime.date, datetime.time]
    ) -> bool:

        begin = self.begin or datetime.datetime.min
        end = self.end or datetime.datetime.max

        if isinstance(value, datetime.time):
            return begin.time() <= value <= end.time()
        elif isinstance(value, datetime.datetime):
            return begin <= value <= end
        elif isinstance(value, datetime.date):
            return begin.date() <= value <= end.date()

        raise TypeError("Can't compare TimeRange to %s", type(value).__name__)

    def __eq__(self, other):
        try:
            return (
                self.__slots__ == other.__slots__
                and self.begin == other.begin
                and self.end == other.end
            )
        except AttributeError:
            return False

    def range(self, frame: str):
        """
        Generator yielding datetime values within a particular frame: hours/days/weeks/etc

        :param frame:
        :return:
        @throws:        OverflowError in open-end datetime case
        """

        current = self.begin or datetime.datetime.now()
        end = self.end or datetime.datetime.max

        while current <= end:
            yield current
            current += datetime.timedelta(**{frame: 1})

    def __str__(self):
        return f'<TimeRange begin="{self.begin}" end="{self.end}">'


class TimeSet:
    """
    Recurring DateTimeV2 set
    """

    __slots__ = ("timex", "tz")

    timex: str  # Unparsed (original) value from NLU
    tz: datetime.tzinfo  # Time zone

    MAX: datetime.datetime = (
        datetime.datetime.max
    )  # Maximum allowed date/time to stop iteration

    def __init__(self, timex: str, tz=tzutc()):
        if not isinstance(tz, datetime.tzinfo):
            tz = gettz(tz)

        self.tz = tz
        self.timex = timex

    def range(self, count: int = None, until: str = None):
        until_date: Optional[datetime.datetime]
        if until is not None:
            if count is not None:
                logger.debug(
                    '"count" and "until" should not be used together, setting "count" to "None"'
                )
                count = None
            until_date = parser.parse(until)
            if until_date.tzinfo is None:
                until_date = until_date.replace(tzinfo=self.tz)
        else:
            until_date = None if count else self.MAX.replace(tzinfo=self.tz)

        try:

            rule = _parse_timex(self.timex)
            if rule["dtstart"].tzinfo is None:
                rule["dtstart"] = rule["dtstart"].replace(tzinfo=self.tz)

            if rule["dtstart"] < datetime.datetime.now(self.tz):
                rule["dtstart"] += datetime.timedelta(days=1)

            return rrule.rrule(**{**rule, **dict(count=count, until=until_date)})

        except (TypeError, ValueError) as ex:
            raise ValueError(f'Could not parse timex value: "{self.timex}", {ex}')

    def __str__(self):
        return f'<TimeSet timex="{self.timex}" tz="{self.tz}">'


class AttributeV2(CamelModel, Generic[T]):
    """
    Attribute V2: indicates the nested and overlapping entities

        Sample usage:

        >>>import datetime
        >>>from skill_sdk import skill, responses
        >>>from skill_sdk.intents import AttributeV2

        >>>@skill.intent_handler('Intent')
        >>>def handler(date: AttributeV2[str]) -> responses.Response:
        >>>    ...

    """

    # Attribute ID (int32)
    id: int

    # Attribute value
    value: T

    # List of attribute IDs this attribute is nested in
    nested_in: List[int] = []

    # List of attribute IDs this attribute overlaps with
    overlaps_with: List[int] = []

    # Extra information for skills. e.g. literal value
    extras: Optional[Dict[Text, Text]]

    class Config:
        """Sample values for Swagger UI"""

        schema_extra = {
            "example": {
                "id": 1,
                "value": "value",
                "nested_in": [],
                "overlaps_with": [],
                "extras": {},
            }
        }

    @root_validator(pre=True)
    def remove_none_values(cls, values: Dict):  # pylint: disable=E0213
        return {k: v for k, v in values.items() if v is not None}

    def __init__(
        self,
        value: Union[Dict, "AttributeV2", Text],
        mapping: Callable[[Any], Any] = None,
        **data: Any,
    ) -> None:
        """
        Create AttributeV2 object optionally converting the value to required type

        :param value:   Value dictionary
        :param mapping: Conversion function
        """
        if isinstance(value, type(self)):
            __value = value.dict()
        elif isinstance(value, dict):
            __value = value.copy()
        else:
            __value = dict(value=value, **data)

        if mapping:
            try:
                __value["value"] = mapping(__value["value"])
            except (KeyError, TypeError):
                pass

        super().__init__(**__value)


def rank(value: str) -> int:
    """
    Convert GENERIC_ORDER NLU concept to an index value.
    The GENERIC_ORDER concept return a string that can be "max", "min", "succ", "prec", "0", "1", ...<br>
        "max" -> -1<br>
        "min" -> 0<br>
        "1" -> 0<br>
        "2" -> 1<br>
        ...<br>
    The "succ" value is not handled and should not be used in the intent samples.<br>
    Please see https://developer.nuance.com/mix/documentation/nlu/#nuance_generic_order

    :param value:   GENERIC_ORDER concept value.
    :return: index
    """
    _ = {
        "min": 0,
        "max": -1,
        "prec": -2,
    }
    try:
        return _[value]
    except KeyError:
        return int(value) - 1


def converter(to_type) -> Callable[[str], Any]:
    """
    Returns conversion function

    :param to_type: type or callable
    :return:
    """
    _ = {
        datetime.timedelta: to_timedelta,
        datetime.datetime: to_datetime,
        datetime.date: to_date,
        datetime.time: to_time,
        bool: on_off_to_boolean,
    }

    conversion_func = to_type if callable(to_type) else lambda a: a
    return _.get(to_type, conversion_func)


def convert(value: str, to_type=None):
    """
    Convert value to type defined in conversion table or using a conversion function provided in `to_type` param
    returns converted value or the value itself if conversion function not defined

    :param value:   value to convert
    :param to_type: type or callable
    :return:
    """
    return converter(to_type)(value)


@to_datetime.register(datetime.date)
def date_to_datetime(value: datetime.date) -> datetime.datetime:
    """Date to datetime with time set to 00:00"""
    return datetime.datetime.combine(value, datetime.time(0, 0))


@to_datetime.register(datetime.time)
def time_to_datetime(value: datetime.time) -> datetime.datetime:
    """Time to datetime with date set to today"""
    return datetime.datetime.combine(datetime.datetime.now().date(), value)


@to_datetime.register(list)
def list_to_datetime(entities: List) -> datetime.datetime:
    """Parse first list element and return datetime"""
    value = get_entity(entities)
    return to_datetime(value) if value else datetime.datetime.min


@to_date.register(datetime.datetime)
def datetime_to_date(value: datetime.datetime) -> datetime.date:
    """Parse datetime and return date component"""
    return to_datetime(value).date() if value else datetime.date.min


@to_date.register(list)
def list_to_date(entities: List) -> datetime.date:
    """Parse first list element and return date component"""
    value = get_entity(entities)
    return to_date(value) if value else datetime.date.min


@to_time.register(datetime.datetime)
def datetime_to_time(value: datetime.datetime) -> datetime.time:
    """Parse datetime and return time component"""
    return value.time()


@to_time.register(list)
def list_to_time(entities: List) -> datetime.time:
    """Parse first list element and return time component"""
    value = get_entity(entities)
    return to_time(value) if value else datetime.time.min


WEEK_TIMEX = re.compile(
    r"^(?:(?P<year>[X\d]+)-)?" r"W(?:(?P<week>[X\d]+)-)?" r"(?:(?P<day>\d+))?$"
)


def _parse_timex(timex: str) -> Dict:
    """
    Parse "timex" value from NLU
    Currently covers the following cases:

        - "every day at ...":           "T%HH:%MM"
        - "every Monday at ...":        "XXXX-WXX-1T%HH:%MM"
        - "every Monday to Wednesday":  "(XXXX-WXX-1,XXXX-WXX-3,P2D)"
        - "every Monday ... to ...":    "(XXXX-WXX-1T14,XXXX-WXX-1T18,PT4H)"

    :param timex:
    :return:
    """
    weekday = None
    freq = rrule.DAILY

    timex_tuple = _parse_timex_tuple(timex)
    if timex_tuple:
        return timex_tuple

    if "T" in timex:
        date_part, time_part = timex.split("T")
        # if Timex ends with 'Txx', we want to add ':00' to specify that it's a time expression
        if ":" not in time_part:
            time_part = f"{time_part}:00"
    else:
        date_part = timex
        time_part = "00:00"

    if "W" in date_part:
        result = WEEK_TIMEX.search(date_part)
        if result:
            year, week, day = result.groupdict().values()
            freq = rrule.WEEKLY
            weekday = int(day) - 1  # rrule says Monday is 0
            date_part = ""

    return {
        "dtstart": parser.parse("T".join((date_part, time_part))),
        "freq": freq,
        "byweekday": weekday,
    }


def _parse_timex_tuple(timex: str) -> Optional[Dict]:
    """
    Try to parse duration tuple: (begin,end,duration)

    :param timex:
    :return:
    """
    try:
        begin, end, duration = timex.strip("()").split(",")
        start_args = _parse_timex(begin)
        end_args = _parse_timex(end)
        delta = isodate.parse_duration(duration)
        if delta.days:
            return {
                "dtstart": start_args["dtstart"],
                "freq": rrule.WEEKLY,
                "byweekday": range(start_args["byweekday"], end_args["byweekday"] + 1),
            }
        elif delta.seconds:
            return {
                "dtstart": start_args["dtstart"],
                "freq": rrule.WEEKLY,
                "byweekday": start_args["byweekday"],
                # For weekly recurrence, time part of `dtstart` is ignored
                "byhour": range(
                    start_args["dtstart"].hour, end_args["dtstart"].hour + 1
                ),
            }
        else:
            raise NotImplementedError(f"Cannot parse value {timex} (yet)")
    except ValueError:
        return None



from enum import Enum
from typing import Callable, List, Dict, Optional, Any, Text
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from fastapi import status
from skill_sdk.utils.util import CamelModel


class SessionRequestDto(CamelModel):
    new_session: bool
    id: str
    attributes: Dict[str, str]


class SkillContextAttributeValueDto(CamelModel):
    id: int
    value: str
    nested_in: List[int]
    overlaps_with: List[int]
    extras: Dict[str, str]


class SkillContextDto(CamelModel):
    intent: str
    skill_id: str
    attributes: Optional[Dict[str, List[str]]]
    attributesV2: Optional[Dict[str, List[SkillContextAttributeValueDto]]]
    tokens: Dict[str, str]
    locale: str
    configuration: Dict[str, List[str]]
    user_profile_config: str
    client_type_name: str


class SkillFitnessInitiateJudgement(CamelModel):
    context: SkillContextDto
    session: SessionRequestDto
    spi_version: str


class SkillFitContentMap(CamelModel):
    skillFitConfidence: float


class SkillAssessmentResultType(Text, Enum):
    NOT_IMPLEMENTED = "501"
    CONNECTION_FAILED = "502"
    TIMEOUT = "408"
    OK = "200"


class SkillFitAssessmentObject(CamelModel):
    result_type: SkillAssessmentResultType
    content_version: str
    content_map: Dict[str, str]


class SkillFitAssessmentSingleton(object):
    _instance = None
    _skill_fit_assess_func: Callable

    def __init__(self):
        if SkillFitAssessmentSingleton._instance is None:
            raise RuntimeError('Call SkillFitAssessmentSingleton.instance() instead')
        else:
            SkillFitAssessmentSingleton._instance = self

    def set_response_implementation(self, func_name):
        self._skill_fit_assess_func = func_name

    def get_response_implementation(self, r: SkillFitnessInitiateJudgement):
        response_object = self._skill_fit_assess_func(r)
        return JSONResponse(status_code=status.HTTP_200_OK,
                            content=dict(resultType=response_object.result_type,
                                         contentMap=jsonable_encoder(response_object.content_map),
                                         contentVersion=response_object.content_version))

    def return_default_response_implementation(self, r: SkillFitnessInitiateJudgement):
        return SkillFitAssessmentObject(result_type=SkillAssessmentResultType.NOT_IMPLEMENTED, content_version=1.0,
                                        content_map=dict(SkillFitContentMap(skillFitConfidence=0.0)))

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            cls._skill_fit_assess_func = cls.return_default_response_implementation

        return cls._instance



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Create application instance from config files"""

import os
import re
import uuid
import logging
from pathlib import Path
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Text, Tuple, Union
from configparser import BasicInterpolation, ConfigParser, SectionProxy

from pydantic import BaseSettings, Extra, fields, validator
from pydantic.env_settings import SettingsSourceCallable


logger = logging.getLogger(__name__)

ENV_VAR_TEMPLATE = re.compile(r"^\${(.*)}$")

# Skill config default file name
SKILL_CONFIG_FILE = "skill.conf"

# Dotenv variables default file name
DOTENV_FILE = ".env"


class EnvVarInterpolation(BasicInterpolation):
    """
    Interpolation to expand environment variables in the format:

    [section]
    key = ${ENV_VAR:default}

    """

    def before_get(self, parser, section, option, value, defaults):
        """
        If value matches the template (${ENV_VAR:default}),
        extract the value from environment variable,
        or set it to default

        :param parser:
        :param section:
        :param option:
        :param value:
        :param defaults:
        :return:
        """
        match = ENV_VAR_TEMPLATE.match(value)
        if not match:
            return os.path.expandvars(value)

        match = match.group(1)
        env_var, default = match.split(":", 1) if ":" in match else (match, None)
        value = os.getenv(env_var)
        if value:
            logger.debug("Read %s from environment: %s", env_var, value)
        elif default:
            logger.debug("%s is empty, setting to default: %s", env_var, default)
        return value or default


def get_skill_config_file(config: Union[Path, Text] = None) -> Text:
    """
    Check if skill config file exists

    :param config:
    :return:
    """
    config_file = (
        config if config is not None else os.getenv("CONFIG_FILE", SKILL_CONFIG_FILE)
    )

    try:
        with open(config_file):
            logger.info("Using config: %s", config_file)

    except FileNotFoundError as ex:
        if config:
            raise RuntimeError from ex
        logger.error("File not found: %s", repr(config_file))

    return str(config_file)


def load_additional() -> List[Text]:
    """
    Load additional configuration files

    :return:
    """
    config: List[Text] = []
    additional_locations = os.getenv("CONFIG_ADDITIONAL_LOCATION")

    if not additional_locations:
        logger.debug("CONFIG_ADDITIONAL_LOCATION is not set.")
        return config

    paths = [Path(location).absolute() for location in additional_locations.split(",")]
    logger.debug("Will search configuration in %s", repr(paths))

    for path in paths:
        if path.exists():
            logger.debug("Loading additional configuration files from %s", repr(path))
            for _ in path.iterdir():
                if _.is_file() and _.suffix in (".conf", ".cnf"):
                    config += [str(_)]

    return config


def read_config(path: List[Text]) -> ConfigParser:
    """
    Read configuration from files

    :param path:    path to the configuration file
    :return:        configuration dictionary
    """

    logger.debug("Loading configuration from %s", repr(path))

    config = ConfigParser(interpolation=EnvVarInterpolation())
    paths = config.read(path)

    logger.debug("Loaded from %s", repr(paths))

    return config


def init_config(config: Union[Dict, Path, Text, None]) -> ConfigParser:
    """
    Initialize skill configuration:

        1. Defaults from SDK
        2. Additional locations
        3. "skill.conf"

    :param config:
    :return:
    """

    # Merge additional configs
    config_paths = load_additional()

    if not isinstance(config, dict):
        config_paths += [get_skill_config_file(config)]

    parser = read_config(config_paths)

    if isinstance(config, dict):
        logger.debug("Creating app from dictionary: %s", repr(config))
        parser.read_dict(config)

    return parser


def clean_section(d: SectionProxy, **kwargs) -> Dict[Text, Any]:
    """
    Helper to clean a ConfigParser's section and return as dictionary
        convert boolean values to bools, integer values to ints, float values to floats

    :param d:
    :param kwargs:
    :return:
    """

    def get(option, chain: Iterator):
        """
        Consequentially apply operations from the chain on the value provided by `option` parameter,
        if operation raises an exception, try the next one

        :param option:
        :param chain:
        :return:
        """
        try:
            return next(chain)(option)
        except (AttributeError, TypeError, ValueError):
            return get(option, chain)

    return {
        o: kwargs[o]
        if o in kwargs
        else get(o, iter((d.getint, d.getfloat, d.getboolean, d.get)))
        for o, v in {**dict(d.items()), **kwargs}.items()
    }


#
# Skill configuration settings
#


class FormatType(str, Enum):
    """Types of log format: human-readable or machine-readable"""

    GELF = "gelf"
    HUMAN = "human"


class Settings(BaseSettings):
    """Skill settings"""

    # Default values for skill name/title/description and version
    SKILL_NAME: Text = "skill-noname"
    SKILL_TITLE: Text = "Magenta Skill SDK Python"
    SKILL_DESCRIPTION: Text = "Magenta Voice Skill SDK for Python"
    SKILL_VERSION: Text = "1"

    # API base, if None, will default to "/v{settings.SKILL_VERSION}/{settings.SKILL_NAME}"
    API_BASE: Optional[Text] = None

    SKILL_DEBUG: bool = False

    # Basic auth: cvi/API key
    SKILL_API_USER: Text = "cvi"
    SKILL_API_KEY: Text = uuid.uuid1().__str__()

    @validator("SKILL_API_KEY", pre=True, always=True)
    def generate(cls, v):  # pylint: disable=E0213
        return v or uuid.uuid1().__str__()

    # Default HTTP port
    HTTP_PORT: int = 4242

    # Health endpoints for k8s
    K8S_READINESS: Text = "/k8s/readiness"
    K8S_LIVENESS: Text = "/k8s/liveness"

    # Prometheus metrics scraper endpoint
    PROMETHEUS_ENDPOINT: Text = "/prometheus"

    # Skill assessment endpoint
    SKILL_FIT_ASSESSMENT_ENDPOINT: Text = "/skill-fit-assessment"

    # Default request time-out value in seconds:
    # used from built-in httpx client
    REQUESTS_TIMEOUT: float = 5

    #
    # Logging
    #

    # Default log level: WARNING
    LOG_LEVEL: int = logging.WARNING

    # Default log format: GELF
    LOG_FORMAT: FormatType = FormatType.GELF

    @validator("LOG_LEVEL", pre=True, allow_reuse=True)
    def to_int(cls, v):  # pylint: disable=E0213
        return v if isinstance(v, int) else logging.getLevelName(v)

    # Maximal length of a string in the log
    LOG_ENTRY_MAX_STRING: int = 150

    # JSON-formatted list of CORS origins: requests from dev and prod UI
    BACKEND_CORS_ORIGINS: List[Text] = [
        "http://localhost:8080",
        "http://127.0.0.1:8080",
        "http://localhost:4242",
        "http://127.0.0.1:4242",
    ]

    # The JWT secret that is used to verify the cvi service-token
    CVI_SERVICE_TOKEN_SECRET: str = "${CVI_SERVICE_TOKEN_SECRET:False}"

    def debug(self) -> bool:
        """
        Tell if skill is in "debug" mode

        :return:
        """
        return self.SKILL_DEBUG

    def app_config(self) -> Dict[Text, Any]:
        """
        This section values will be used to initialize FastAPI object

        :return:
        """
        return dict(
            title=self.SKILL_NAME,
            version=self.SKILL_VERSION,
            description=self.SKILL_DESCRIPTION,
            debug=self.SKILL_DEBUG,
        )

    def http_config(self) -> Dict[Text, Any]:
        """
        These values will be forwarded as keyword arguments to HTTP server

        :return:
        """
        return dict(port=self.HTTP_PORT)

    @classmethod
    def add_fields(cls, **field_definitions: Any):
        """
        Add new fields to the model

        :param field_definitions:
        :return:
        """
        new_fields: Dict[str, fields.ModelField] = {}
        new_annotations: Dict[str, Optional[type]] = {}

        for f_name, f_def in field_definitions.items():
            f_annotation, f_value = type(f_def), f_def

            # if field is not defined, add new field with annotation to the model
            if f_name not in cls.__fields__:
                if f_annotation:
                    new_annotations[f_name] = f_annotation

                new_fields[f_name] = fields.ModelField.infer(
                    name=f_name,
                    value=f_value,
                    annotation=f_annotation,
                    class_validators=None,
                    config=cls.__config__,  # type: ignore
                )

        cls.__fields__.update(new_fields)
        cls.__annotations__.update(new_annotations)

    def reload(
        self, conf_file: Union[Dict, Path, Text, None] = None, **values
    ) -> "Settings":
        """
        Reload values from new config file:

            creates a new Settings instance,
            and performs in-place update of this singleton

        :param conf_file:
        :param values:
        :return:
        """
        Settings.Config.conf_file = conf_file
        for key, value in Settings(**values):
            setattr(self, key, value)

        return self

    class Config:
        """
        Config values priority:

            1. Arguments passed to the Settings class constructor.
            2. Environment variables.
            3. Variables loaded from a dotenv (.env) file.
            4. Variable values from skill.conf file.
            5. Variables loaded from the secrets directory.

        """

        extra = Extra.allow
        env_file = DOTENV_FILE
        conf_file: Union[Dict, Path, Text, None] = SKILL_CONFIG_FILE

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            """
            Override the parent to insert "skill_conf_settings" values

            :param init_settings:
            :param env_settings:
            :param file_secret_settings:
            :return:
            """
            return (
                init_settings,
                env_settings,
                skill_conf_settings,
                file_secret_settings,
            )


def skill_conf_settings(_settings: BaseSettings) -> Dict[str, Any]:
    """
    Read configuration setting in ConfigParser format from "skill.conf":

        replace dashes with underscores in section names and configuration keys,
        join names and keys with underscore ("_") and convert to upper case

        So that a following config:

        ```ini
        [section]
        key = Value
        ```

        will be available as:

        >>> settings.SECTION_KEY
        >>> "Value"

    :param _settings:
    :return:
    """

    try:
        c: ConfigParser = init_config(Settings.Config.conf_file)

        d: Dict[Text, Any] = {
            "_".join((_make_key(section), _make_key(field))): value
            for section in c.sections()
            for field, value in clean_section(c[section]).items()
        }

        if isinstance(_settings, Settings):
            _settings.add_fields(**d)

        return d

    except RuntimeError:
        return {}


def _make_key(string: Text) -> Text:
    """
    Helper to convert section/value keys to env vars format:

        upper case and replace dashes with underscores

    :param string:
    :return:
    """
    return string.upper().replace("-", "_")


settings = Settings()



#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors /
# copyright owners license this file to you under the MIT
# License (the "License"); you may not use this file
# except in compliance with the License.
# You may obtain a copy of the License at
#
# https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""Magenta Voice Skill SDK for Python"""

from .__version__ import __version__

from skill_sdk.skill import init_app, Skill

from skill_sdk.responses import (
    Card,
    CardAction,
    Response,
    ResponseType,
    Reprompt,
    SkillInfoResponse,
    SkillInvokeResponse,
    ask,
    ask_freetext,
    tell,
)

from skill_sdk.intents import Request



#
#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#

"""Companion app (cApp) cards"""

from enum import Enum
from warnings import warn
from typing import List, Optional, Text, Union
from skill_sdk.utils.util import CamelModel

# Supported card version
CARD_VERSION = 3

# The only card type available: "GENERIC_DEFAULT"
GENERIC_DEFAULT = "GENERIC_DEFAULT"


class CardAction(Text, Enum):
    """
    Card action link can be either one of internal "deep links" (enumerated below)
        or any external "http/https" URL

    """

    # Present skill details view
    INTERNAL_SKILLS = "internal://deeplink/skills"

    # Present overview of all devices
    INTERNAL_OVERVIEW = "internal://deeplink/speakeroverview"

    # Present device details page of the one device that was spoken into to generate this card
    INTERNAL_DETAILS = "internal://deeplink/speakerdetails"

    # Present feedback page in the app
    INTERNAL_FEEDBACK = "internal://deeplink/feedback"

    # Link to the news section of the app
    INTERNAL_NEWS = "internal://deeplink/news"

    # Present full text of the response in an overlay
    INTERNAL_RESPONSE_TEXT = "internal://showResponseText"

    # Initiate a call to the given phone number.
    INTERNAL_CALL = "internal://deeplink/call/{number}"

    # Open a specified app or the App Store if the app is not installed
    INTERNAL_OPEN_APP = (
        "internal://deeplink/openapp?"
        "aos={aos_package_name}&"
        "iosScheme={ios_url_scheme}&"
        "iosAppStoreId={ios_app_store_id}"
    )


class ListItem(CamelModel):
    """
    List item in Card's list sections
    """

    # List item text
    item_text: Text

    # List item action
    item_action: Optional[Union[CardAction, Text]] = None

    # List item bullet point replacement
    item_icon_url: Optional[Text] = None

    @property
    def title(self):
        """**DEPRECATED**: Item title"""
        warn(
            "ListItem.title is deprecated. Use ListItem.item_text instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return None

    @property
    def icon_url(self):
        """**DEPRECATED**: Item icon URL"""
        warn(
            "ListItem.icon_url is deprecated. Use ListItem.item_icon_url instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return None


class ListSection(CamelModel):
    """
    List section in a Card
    """

    title: Optional[Text] = None
    items: List[ListItem] = []

    def __init__(self, title: Text = None, *, items: List[ListItem] = None):
        super().__init__(title=title, items=items or [])

    def with_list_item(
        self,
        item_text: Union[ListItem, Text],
        item_action: Text = None,
        item_icon_url: Text = None,
    ):
        list_item: ListItem = (
            item_text
            if isinstance(item_text, ListItem)
            else ListItem(
                item_text=item_text,
                item_action=item_action,
                item_icon_url=item_icon_url,
            )
        )
        return self.copy(update=dict(items=self.items + [list_item]))


class CardData(CamelModel):
    """Companion app card data"""

    # Card's icon URL
    icon_url: Optional[Text] = None

    # Card title
    title_text: Optional[Text] = None

    # Card subtitle
    type_description: Optional[Text] = None

    # Card's image URL
    image_url: Optional[Text] = None

    # Prominent text: increased font size, displayed below the image (1 line maximum)
    prominent_text: Optional[Text] = None

    # Prominent text action: uri/url action, triggered when tapping prominent text
    action_prominent_text: Optional[Text] = None

    # Actual card text
    text: Optional[Text] = None

    # Sub-text: decreased font size, displayed below the card text
    # the first 4 lines displayed, the rest is hidden under "Show More"
    sub_text: Optional[Text] = None

    # Optional audio media URL, displayed as media player progress bar
    media_url: Optional[Text] = None

    #
    # Card can have a bullet-like list:
    #   a list section contains items and an optional title
    #   an item has a title and optional iconUrl
    #
    list_sections: Optional[List[ListSection]] = None

    def with_list_section(
        self,
        title: Text,
        items: List[ListItem] = None,
    ):
        list_sections = self.list_sections or []
        return self.copy(
            update=dict(list_sections=list_sections + [ListSection(title, items=items)])
        )

    #
    # **DEPRECATED**: Properties from Card v2.0
    #

    @property
    def action(self):
        """**DEPRECATED**: Action URL linked to the action text"""
        warn(
            "Card.action is deprecated: use ListItem.item_action",
            DeprecationWarning,
            stacklevel=2,
        )
        return None

    @property
    def action_text(self):
        """**DEPRECATED**: Action text"""
        warn(
            "Card.action_text is deprecated: use ListItem.item_action",
            DeprecationWarning,
            stacklevel=2,
        )
        return None


class Card(CamelModel):
    """Card for the companion app"""

    type: Text = GENERIC_DEFAULT
    version: int = CARD_VERSION

    data: CardData

    def __init__(
        self,
        data: CardData = None,
        *,
        icon_url: Text = None,
        title_text: Text = None,
        type_description: Text = None,
        image_url: Text = None,
        prominent_text: Text = None,
        action_prominent_text: Text = None,
        text: Text = None,
        sub_text: Text = None,
        media_url: Text = None,
        list_sections: List[ListSection] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            data=data
            or CardData(
                icon_url=icon_url,
                title_text=title_text,
                type_description=type_description,
                image_url=image_url,
                prominent_text=prominent_text,
                action_prominent_text=action_prominent_text,
                text=text,
                sub_text=sub_text,
                media_url=media_url,
                list_sections=list_sections,
            )
        )

    def __getattr__(self, attr):
        return getattr(self.data, attr)

    def with_list_section(
        self,
        title: Text,
        items: List[ListItem] = None,
    ):
        return self.copy(
            update=dict(data=self.data.with_list_section(title, items=items))
        )

    def with_action(
        self,
        item_text: Text,
        item_action: Union[CardAction, Text],
        item_icon_url: Text = None,
    ) -> "Card":
        """
        Add action to card. This function left for backward compatibility:
        in v3.0 the action item has been removed from cApp card,
        card actions must be defined as action list items within listSections

        This method creates and returns new card,
        replacing listSections field with a single section containing one list action item

        **WARNING**: this will replace existing list section items!

        @param item_text:
        @param item_action:
        @param item_icon_url:

        @return:
        """
        data = self.data.copy(
            update=dict(
                list_sections=[
                    ListSection().with_list_item(
                        item_text=item_text,
                        item_action=item_action,
                        item_icon_url=item_icon_url,
                    )
                ],
            )
        )
        return self.copy(update=dict(data=data))



#
# voice-skill-sdk
#
# (C) 2021, YOUR_NAME (YOUR COMPANY)
#
# This file is distributed under the terms of the MIT license.
# For details see the file LICENSE in the top directory.
#
#

"""Your intent handlers"""

from skill_sdk import Response, skill, tell
from skill_sdk.i18n import _


@skill.intent_handler("SMALLTALK__GREETINGS")
async def handler() -> Response:
    """
    A handler of SMALLTALK__GREETINGS intent.

        SMALLTALK__GREETINGS intent is activated when user says 'Hello'
        returns translated 'Hello' greeting

    :return:        Response
    """

    # Get a translated message
    msg = _("HELLOAPP_HELLO")

    # Create a simple response
    response = tell(msg)

    # Return the response
    return response
